"""
Input set generators for LAMMPS.
This InputSet and InputGenerator implementation is based on templates and is not intended to be very flexible.
For instance, pymatgen will not detect whether a given variable should be adapted based on others
(e.g., the number of steps from the temperature), it will not check for convergence nor will it actually run LAMMPS.
For additional flexibility and automation, use the atomate2-lammps implementation
(https://github.com/Matgenix/atomate2-lammps).
"""

from __future__ import annotations

import os
import warnings
from dataclasses import dataclass, field
from pathlib import Path
from string import Template
from typing import TYPE_CHECKING

import numpy as np
from monty.io import zopen
from monty.json import MSONable

from pymatgen.core import Lattice, Molecule, Structure
from pymatgen.io.core import InputGenerator
from pymatgen.io.lammps.data import CombinedData, LammpsBox, LammpsData, LammpsForceField
from pymatgen.io.lammps.inputs import LammpsInputFile
from pymatgen.io.lammps.sets import LammpsInputSet
from pymatgen.util.typing import PathLike

if TYPE_CHECKING:
    from typing import Literal

    from numpy.typing import ArrayLike

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
TEMPLATE_DIR = f"{MODULE_DIR}/templates"

_COMMON_LAMMPS_SETTINGS = {
    "dimension": 3,
    "pair_style": "lj/cut 10.0",
    "thermo": 100,
    "start_temp": 300.0,
    "end_temp": 300.0,
    "start_pressure": 0.0,
    "end_pressure": 0.0,
    "log_interval": 100,
    "traj_interval": 100,
    "ensemble": "nvt",
    "thermostat": "nose-hoover",
    "barostat": None,
    "nsteps": 1000,
    "restart": "",
    "tol": 1e-6,
    "min_style": "cg",
}

_BASE_LAMMPS_SETTINGS = {
    "periodic": _COMMON_LAMMPS_SETTINGS
    | {
        "units": "metal",
        "atom_style": "atomic",
        "boundary": ("p", "p", "p"),
        "timestep": 0.001,
        "friction": 0.1,
    },
    "molecular": _COMMON_LAMMPS_SETTINGS
    | {
        "units": "real",
        "atom_style": "full",
        "boundary": ("f", "f", "f"),
        "timestep": 1,
        "friction": 100,
    },
}


FF_STYLE_KEYS = ["pair_style", "bond_style", "angle_style", "dihedral_style", "improper_style", "kspace_style"]
FF_COEFF_KEYS = ["pair_coeff", "bond_coeff", "angle_coeff", "dihedral_coeff", "improper_coeff", "kspace_coeff"]

LAMMPS_DEFINED_TYPES: dict[str, set[str | None]] = {
    "units": {"metal", "lj", "real", "si", "cgs", "electron", "micro", "nano"},
    "atom_style": {"atomic", "angle", "body", "bond", "charge", "electron", "full", "molecular"},
    "boundary": {"p", "f", "s", "m", "fs", "fm"},
    "ensemble": {"nve", "nvt", "npt", "nph", "minimize"},
    "thermostat": {"nose-hoover", "langevin", None},
    "barostat": {"nose-hoover", "berendsen", "langevin", None},
    "min_style": {"cg", "sd", "fire", "hftn", "quickmin", "spin", "spin/cg", "spin/lbfgs"},
}


@dataclass
class LammpsSettings(MSONable):
    """Define schema for LAMMPS convenience input class.

    The idea is to dynamically set the attributes of this class
    based on the LAMMPS input file template and the settings
    dictionary provided by the user. This way, standard inputs
    that do not typically vary (e.g., units, atom style, dimension,
    boundary conditions, styles) can be validated and set
    while allowing the user to specify the rest of the settings in
    a flexible manner (which won't be explicitly validated here).

    Args:
    units : str
        LAMMPS units style.
    atom_style : str
        LAMMPS atom style.
    dimension : int
        LAMMPS box dimension.
    boundary : tuple[str,str,str]
        Box boundary conditions. Each boundary can be 'p', 'f', 's', 'm', 'fs', 'fm'.
    pair_style : str
        FF pair style type. Default is 'lj/cut 10.0'.
    bond_style : str
        FF bond style type.
    angle_style : str
        FF angle style type.
    dihedral_style : str
        FF dihedral style type.
    start_temp : float
        Initial temperature. Default is 300 K. Initial velocities are generated based on this temperature.
    end_temp : float
        Final temperature. Default is 300 K.
    start_pressure : float | list | np.ndarray
        Initial pressure. Default is 0 atm. Can be a single value or a list for anisotropic pressure.
    end_pressure : float | list | np.ndarray
        Final pressure. Default is 0 atm. Can be a single value or a list for anisotropic pressure.
    timestep : float
        Simulation timestep. Default is 0.001 ps for periodic solids and 1fs for molecular systems.
    friction : float
        Thermostat/Barostat friction coefficient. Default is 0.1 ps^-1 for periodic solids
        and 100 fs^-1 for molecular systems.
    log_interval : int
        Interval for logging thermodynamic data (energy, temperature, pressure, volume).
        Default is 100 steps.
    traj_interval : int
        Interval for writing trajectory data (positions, velocities, forces) to a dump file.
        Default is 100 steps.
    ensemble : str
        Simulation ensemble. Default is 'nvt'.
    thermostat : str
        Thermostat type. Default is 'nose-hoover'.
    barostat : str
        Barostat type. Default is 'nose-hoover'.
    nsteps : int
        Number of simulation steps. Default is 1000. This is also the maximum number of
        steps for minimization simulations.
    """

    validate_params: bool = field(default=False)

    def __init__(self, validate_params: bool = False, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        if validate_params:
            self.__post_init__()
        else:
            # Always call __post_init__ for warnings, but skip validation
            self._post_init_warnings_only()

    def _post_init_warnings_only(self) -> None:
        """Run warnings only, no validation."""
        # Always validate restart parameter
        if hasattr(self, "restart") and self.restart and not isinstance(self.restart, str):
            raise ValueError(
                f"restart should be the path to the restart file from the previous run, got {self.restart}."
            )

        if (
            hasattr(self, "ensemble")
            and self.ensemble not in ["nve", "minimize"]
            and hasattr(self, "friction")
            and hasattr(self, "timestep")
            and (getattr(self, "thermostat", None) or getattr(self, "barostat", None))
            and self.friction < self.timestep
        ):
            warnings.warn(
                f"Friction ({self.friction}) is smaller than the timestep ({self.timestep}). "
                "This may lead to numerical instability.",
                stacklevel=2,
            )

        if hasattr(self, "ensemble") and self.ensemble == "minimize" and hasattr(self, "tol") and self.tol > 1e-4:
            warnings.warn(
                f"Tolerance for minimization ({self.tol}) is larger than 1e-4. This may lead to inaccurate results.",
                stacklevel=2,
            )

    def __post_init__(self) -> None:
        """Validate input values."""

        for attr, accept_vals in LAMMPS_DEFINED_TYPES.items():
            curr_val = getattr(self, attr, None)
            if curr_val is not None:  # Only validate if the attribute exists
                if isinstance(curr_val, list | tuple):
                    is_ok = all(v in accept_vals for v in curr_val)
                else:
                    is_ok = curr_val in accept_vals
                if not is_ok:
                    raise ValueError(f"Error validating key {attr}: set to {curr_val}, should be one of {accept_vals}.")

        if hasattr(self, "restart") and self.restart and not isinstance(self.restart, str):
            raise ValueError(
                f"restart should be the path to the restart file from the previous run, got {self.restart}."
            )

        if hasattr(self, "ensemble") and self.ensemble not in ["nve", "minimize"]:
            if (
                hasattr(self, "start_pressure")
                and isinstance(self.start_pressure, (list, np.ndarray))
                and len(self.start_pressure) != 3
            ):
                raise ValueError(f"start_pressure should be a list of 3 values, got {self.start_pressure}.")
            if (
                hasattr(self, "end_pressure")
                and isinstance(self.end_pressure, (list, np.ndarray))
                and len(self.end_pressure) != 3
            ):
                raise ValueError(f"end_pressure should be a list of 3 values, got {self.end_pressure}.")

            if (
                hasattr(self, "friction")
                and hasattr(self, "timestep")
                and (getattr(self, "thermostat", None) or getattr(self, "barostat", None))
                and self.friction < self.timestep
            ):
                warnings.warn(
                    f"Friction ({self.friction}) is smaller than the timestep ({self.timestep}). "
                    "This may lead to numerical instability.",
                    stacklevel=2,
                )

        if hasattr(self, "ensemble") and self.ensemble == "minimize":
            if hasattr(self, "nsteps") and self.nsteps < 1:
                raise ValueError(f"nsteps should be greater than 0 for minimization simulations, got {self.nsteps}.")
            if hasattr(self, "tol") and self.tol > 1e-4:
                warnings.warn(
                    f"Tolerance for minimization ({self.tol}) is larger than 1e-4. "
                    "This may lead to inaccurate results.",
                    stacklevel=2,
                )

    def update(self, updates: dict) -> None:
        for key, value in updates.items():
            setattr(self, key, value)

    def as_dict(self) -> dict:
        """Convert to dictionary for serialization."""
        # Just return the plain dict representation without MSONable metadata
        # This makes it serializable as a regular dict
        return self.__dict__

    @classmethod
    def from_dict(cls, d: dict) -> LammpsSettings:
        """Create LammpsSettings from dictionary."""
        # Remove MSONable metadata if present
        clean_dict = {k: v for k, v in d.items() if not k.startswith("@")}
        return cls(**clean_dict)


@dataclass
class BaseLammpsSetGenerator(InputGenerator):
    """
    Base class for generating LAMMPS input sets.

    Args:
        inputfile : LammpsInputFile | str | Path
            Premade input file for the LAMMPS simulation.
            Useful if the user wants to use a custom input file
            (to make use of Lammps' flexibility).
            Default format based on the md.template file in the templates directory.
        settings : dict | LammpsInputSettings
            Settings for the LAMMPS simulation. Default settings are given in the
            _BASE_LAMMPS_SETTINGS object in 'metal' units for reference.
        force_field : Path | dict | ForceField
            Force field file or dictionary containing the force field parameters.
            Default is None.
        calc_type : str
            Type of calculation to be performed by LAMMPS.
        keep_stages : bool
            Whether to keep the stages of the input file or not. Default is True.
        override_updates : bool
            Whether to override the updates to the input file, i.e.,
            keep the input file as is. Default is False.
    """

    inputfile: LammpsInputFile | str | PathLike = field(default=None)
    settings: dict | LammpsSettings = field(default_factory=dict)
    force_field: Path | dict | LammpsForceField | None = field(default=None)
    data_type: Literal["periodic", "molecular"] = field(default="periodic")
    calc_type: str = field(default="lammps")
    include_defaults: bool = field(default=True)
    validate_params: bool = field(default=True)
    keep_stages: bool = field(default=True)
    override_updates: bool = field(default=False)

    def __post_init__(self):
        """Determine correct template requested by user.

        If not inputfile, we assume the user is going to use the default template for
        one the defined flows (NVT/NVE/NPT, Minimize, MeltQuench, etc.)
        Instead, if the user/another flow specifies a template, we'll use that and an
        optional settings dict with it and not bother with validating the inputs in the
        inputfile due to the amount of flexibility LAMMPS allows
        If another flow is used that defines it's own inputfile template, we'll leave the
        validation of the user's inputs to that flow's init method.

        """
        if self.inputfile is None:
            ensemble = (
                self.settings.get("ensemble", "nvt") if isinstance(self.settings, dict) else self.settings.ensemble
            )

            file: str | None = None
            if ensemble in ["nve", "nvt", "npt", "nph"]:
                file = "md.template"
            elif ensemble == "minimize":
                file = "minimization.template"
            if not file:
                raise ValueError(
                    f"Unknown {ensemble=}; acceptable values are\n{', '.join(LAMMPS_DEFINED_TYPES['ensemble'])}"
                )

            self.inputfile = LammpsInputFile.from_file(os.path.join(TEMPLATE_DIR, file), keep_stages=self.keep_stages)

        if isinstance(self.settings, dict):
            settings = self.settings.copy()
            if self.include_defaults:
                base_settings = _BASE_LAMMPS_SETTINGS[self.data_type].copy()
                settings = base_settings | settings
            self.settings = LammpsSettings(validate_params=self.validate_params, **settings)

        if isinstance(self.force_field, dict):
            self.force_field = LammpsForceField.from_dict(self.force_field)

    def update_settings(
        self, updates: dict, validate_params: bool | None = None, include_defaults: bool | None = None
    ) -> None:
        """
        Update the settings for the LammpsSettings object.
        Args:
            updates : dict
                Dictionary containing the settings to update.
            validate_params : bool
                Whether to validate the parameters. Default is True.
            include_defaults : bool
                Whether to include the default settings. Default is True.
        """
        validate_params = validate_params if isinstance(validate_params, bool) else self.validate_params
        include_defaults = include_defaults if isinstance(include_defaults, bool) else self.include_defaults

        if isinstance(self.settings, LammpsSettings):
            present_settings = self.settings.as_dict().copy()
            if include_defaults:
                base_settings = _BASE_LAMMPS_SETTINGS[self.data_type].copy()
                present_settings = base_settings | present_settings
            for k, v in updates.items():
                present_settings.update({k: v})
            if validate_params:
                present_settings.update({"validate_params": validate_params})
            self.settings = LammpsSettings(**present_settings)
        else:
            if include_defaults:
                base_settings = _BASE_LAMMPS_SETTINGS[self.data_type].copy()
                present_settings = base_settings | updates
            else:
                present_settings = updates
            self.settings.update(present_settings)

    def get_input_set(
        self,
        data: Structure | Molecule | LammpsData | CombinedData,
        additional_data: LammpsData | CombinedData | None = None,
        box_or_lattice: Lattice | LammpsBox | ArrayLike | None = None,
        **kwargs,
    ) -> LammpsInputSet:
        """
        Generate a LAMMPS input set.
        Args:
            data : Structure | Molecule | LammpsData | CombinedData
                Structure, Molecule, LammpsData or CombinedData object containing the atomic structure
                to be simulated.
            additional_data : LammpsData | CombinedData | None
                Additional data to be written to the input set, such as extra data files.
                Default is None.
            data_type : Literal["periodic", "molecular"]
                Type of data provided. Used to determine the atom style and other settings.
                Default is "periodic". If a Molecule is provided, this will be set to "molecular".
            box_or_lattice : Lattice | LammpsBox | ArrayLike | None
                Optional arg needed when data is a Molecule object.
            **kwargs : dict
                Additional keyword arguments to be passed to the LammpsInputSet constructor.
        """

        if not self.force_field:
            warnings.warn(
                "Force field not specified! Ensure you have the correct force field parameters "
                "in the data file/settings or will specify it manually using "
                "maker.input_set_generator.force_field",
                stacklevel=2,
            )

        if isinstance(self.inputfile, PathLike):
            try:
                self.inputfile = LammpsInputFile.from_file(self.inputfile, keep_stages=self.keep_stages)
            except FileNotFoundError:
                try:
                    self.inputfile = LammpsInputFile.from_str(self.inputfile, keep_stages=self.keep_stages)
                except ValueError:
                    raise FileNotFoundError(
                        f"Input file {self.inputfile} not found. It was neither a path "
                        "nor a string repr of the inputfile. Please check your inputs!"
                    )

        # Update data_type based on input
        if isinstance(data, Molecule):
            self.data_type = "molecular"
        self.update_settings(updates={}, validate_params=self.validate_params, include_defaults=self.include_defaults)

        settings_dict = self.settings.as_dict().copy() if isinstance(self.settings, LammpsSettings) else self.settings
        atom_style = settings_dict.get("atom_style", "full")

        # Attempt to read data file and convert to LammpsData object
        if isinstance(data, Path):
            try:
                data = LammpsData.from_file(data, atom_style=atom_style)
            except FileNotFoundError:
                raise FileNotFoundError(f"Data file {data} not found. Please check the path.")

        if isinstance(data, str):
            try:
                data = LammpsData.from_str(data, atom_style=atom_style)
            except ValueError:
                raise ValueError(f"Data file {data} not recognized. Please check the format.")

        species = ""
        if isinstance(data, Structure | Molecule):
            species = " ".join({s.symbol for s in data.species})
            warnings.warn("Structure or Molecule provided, converting to LammpsData object.", stacklevel=2)
            if isinstance(data, Structure):
                data = LammpsData.from_structure(data, atom_style=atom_style)
            elif isinstance(data, Molecule):
                # Provide a default box if none is given
                if box_or_lattice is None:
                    # Create a simple cubic box that's large enough for the molecule
                    from pymatgen.core import Lattice

                    coords = np.array([site.coords for site in data])
                    min_coords = np.min(coords, axis=0)
                    max_coords = np.max(coords, axis=0)
                    box_size = np.max(max_coords - min_coords) + 10.0  # Add 10 Angstrom padding
                    box_or_lattice = Lattice.cubic(box_size)
                data = LammpsData.from_molecule(data, atom_style=atom_style, box_or_lattice=box_or_lattice)

        # Housekeeping to fill up the default settings for the MD template
        settings_dict.update({f"{sys}_flag": "###" for sys in ["nve", "nvt", "npt", "nph", "restart", "extra_data"]})
        settings_dict.update({"read_data_flag": "read_data", "psymm": "iso"})
        # If the ensemble is not 'minimize', we set the read_data_flag to read_data

        # Accounts for the restart file
        if settings_dict.get("read_restart", None):
            print("Restart file provided, setting read_restart flag.")
            settings_dict.update(
                {
                    "read_restart": f"{settings_dict['read_restart']}",
                    "restart_flag": "read_restart",
                    "read_data_flag": "###",
                }
            )

        # Convert start and end pressure to string if they are lists or arrays, and set psymm to accordingly
        if isinstance(settings_dict.get("start_pressure", None), (list, np.ndarray)):
            settings_dict.update(
                {"start_pressure": " ".join(map(str, settings_dict["start_pressure"])), "psymm": "aniso"}
            )
        if isinstance(settings_dict.get("end_pressure", None), (list, np.ndarray)):
            settings_dict.update({"end_pressure": " ".join(map(str, settings_dict["end_pressure"])), "psymm": "aniso"})

        # Loop over the LammpsSettings object and update the settings dictionary
        for attr, val in self.settings.as_dict().items():  # type: ignore[union-attr]
            if attr == "boundary":
                settings_dict.update({"boundary": " ".join(list(val))})

            elif attr == "ensemble":
                settings_dict.update({f"{val}_flag": "fix"})

            elif attr == "thermostat":
                if val == "langevin":
                    settings_dict.update({"nve_flag": "fix", "thermseed": 42, "thermostat": "langevin"})
                if val == "nose-hoover":
                    settings_dict.update({"thermostat": "nvt temp", "thermseed": ""})

            elif attr == "barostat":
                if val == "nose-hoover":
                    settings_dict.update(
                        {
                            "barostat": "npt temp",
                            "start_temp_barostat": settings_dict["start_temp"],
                            "end_temp_barostat": settings_dict["end_temp"],
                            "tfriction_barostat": settings_dict["friction"],
                        }
                    )

                if val in ["berendsen", "langevin"]:
                    settings_dict.update(
                        {
                            "barostat": "langevin" if val == "langevin" else "press/berendsen",
                            "nve_flag": "fix",
                            "nvt_flag": "fix",
                            "start_temp_barostat": "",
                            "end_temp_barostat": "",
                            "tfriction_barostat": "",
                            "thermostat": f"temp/{val}",
                        }
                    )
                    settings_dict.update({"thermoseed": 42 if val == "langevin" else ""})

            elif attr == "friction":
                settings_dict.update({"tfriction": val, "pfriction": val})

            elif val:
                settings_dict.update({attr: val})

        write_data = {}
        # Handle the force field input by writing a separate FF file
        # and making the necessary updates to the settings dict
        if self.force_field:
            FF_string = ""
            if isinstance(self.force_field, str):
                FF_string += self.force_field
                settings_dict.update({f"{ff}_flag": "###" for ff in FF_STYLE_KEYS})

            if isinstance(self.force_field, dict):
                self.force_field = LammpsForceField.from_dict(self.force_field)

            if isinstance(self.force_field, LammpsForceField):
                for key, value in self.force_field.as_dict().items():
                    if key in FF_STYLE_KEYS and value:
                        settings_dict.update({f"{key}": value, f"{key}_flag": f"{key}"})
                    if key in FF_COEFF_KEYS and value:
                        FF_string += f"{key} {value}\n"
                    if key == "species":
                        # makes species specified in FF dict take precedence
                        species = " ".join(value) if isinstance(value, list) else value
                    else:
                        warnings.warn(f"Force field key {key} not recognized, will be ignored.", stacklevel=2)

                for ff_key in FF_STYLE_KEYS:
                    if ff_key not in self.settings.as_dict() or not self.settings.as_dict()[ff_key]:  # type: ignore[union-attr]
                        settings_dict.update({f"{ff_key}_flag": "###"})
                        warnings.warn(
                            f"Force field key {ff_key} not found in the force field dictionary.", stacklevel=2
                        )

            settings_dict.update({"dump_modify_flag": "dump_modify" if species else "###", "species": species})
            write_data.update({"forcefield.lammps": FF_string})

        if additional_data:
            write_data.update({"extra.data": additional_data})
            settings_dict.update({"extra_data_flag": "include"})

        # Replace all variables
        input_str = Template(self.inputfile.get_str()).safe_substitute(**settings_dict)  # type: ignore[union-attr]
        lines = input_str.split("\n")
        # Filter out the lines where the substitution resulted in a line starting with '###'
        filtered_input_str = "\n".join([line for line in lines if not line.lstrip().startswith("###")])
        input_file = LammpsInputFile.from_str(filtered_input_str, keep_stages=self.keep_stages)

        return LammpsInputSet(
            inputfile=input_file, data=data, calc_type=self.calc_type, additional_data=write_data, **kwargs
        )


@dataclass
class BaseLammpsGenerator(InputGenerator):
    r"""
    Base class to generate LAMMPS input sets.
    Uses template files for the input. The variables that can be changed
    in the input template file are those starting with a $ sign, e.g. $nsteps.
    This generic class is specialized for each template in subclasses, e.g. LammpsMinimization.
    You can create a template for your own task following those present in pymatgen/io/lammps/templates.
    The parameters are then replaced based on the values found
    in the settings dictionary that you provide, e.g. `{"nsteps": 1000}`.

    Attributes:
        template: Path (string) to the template file used to create the InputFile for LAMMPS.
        calc_type: Human-readable string used to briefly describe the type of computations performed by LAMMPS.
        settings: Dictionary containing the values of the parameters to replace in the template.
        keep_stages: If True, the string is formatted in a block structure with stage names
        and newlines that differentiate commands in the respective stages of the InputFile.
        If False, stage names are not printed and all commands appear in a single block.

    /!\ This InputSet and InputGenerator implementation is based on templates and is not intended to be very flexible.
    For instance, pymatgen will not detect whether a given variable should be adapted based on others
    (e.g., the number of steps from the temperature), it will not check for convergence nor will it actually run LAMMPS.
    For additional flexibility and automation, use the atomate2-lammps implementation
    (https://github.com/Matgenix/atomate2-lammps).
    """

    inputfile: LammpsInputFile | None = field(default=None)
    template: str | Path | LammpsInputFile = field(default_factory=str)
    data: LammpsData | CombinedData | None = field(default=None)
    settings: dict = field(default_factory=dict)
    calc_type: str = field(default="lammps")
    keep_stages: bool = field(default=True)

    def get_input_set(self, structure: Structure | LammpsData | CombinedData) -> LammpsInputSet:
        """Generate a LammpsInputSet from the structure/data, tailored to the template file."""

        data = (
            LammpsData.from_structure(structure, atom_style=self.settings.get("atom_style", "full"))
            if isinstance(structure, Structure)
            else structure
        )

        # Load the template
        if Path(self.template).is_file():
            with zopen(self.template, mode="rt", encoding="utf-8") as file:
                template_str = file.read()
        elif isinstance(self.template, LammpsInputFile):
            template_str = self.template.get_str()
        else:
            template_str = self.template

        # Replace all variables
        input_str = Template(template_str).safe_substitute(**self.settings)
        # Get LammpsInputFile
        input_file = LammpsInputFile.from_str(input_str, keep_stages=self.keep_stages)

        # Get the LammpsInputSet from the InputFile and data
        return LammpsInputSet(
            inputfile=input_file,
            data=data,
            calc_type=self.calc_type,
            template_file=self.template,
        )


class LammpsMinimization(BaseLammpsGenerator):
    """
    Generator that yields a LammpsInputSet tailored for minimizing the energy of a system by iteratively
    adjusting atom coordinates.
    Example usage:
    ```
    structure = Structure.from_file("mp-149.cif")
    lmp_minimization = LammpsMinimization(units="atomic").get_input_set(structure)
    ```.
    Do not forget to specify the force field, otherwise LAMMPS will not be able to run!
    This InputSet and InputGenerator implementation is based on templates and is not intended to be very flexible.
    For instance, pymatgen will not detect whether a given variable should be adapted based on others
    (e.g., the number of steps from the temperature), it will not check for convergence nor will it actually run LAMMPS.
    For additional flexibility and automation, use the atomate2-lammps implementation
    (https://github.com/Matgenix/atomate2-lammps).
    """

    def __init__(
        self,
        template: str | None = None,
        units: str = "metal",
        atom_style: str = "full",
        dimension: int = 3,
        boundary: str = "p p p",
        read_data: str = "system.data",
        force_field: str = "Unspecified force field!",
        keep_stages: bool = False,
    ) -> None:
        r"""
        Args:
            template: Path (string) to the template file used to create the InputFile for LAMMPS.
            units: units to be used for the LAMMPS calculation (see LAMMPS docs).
            atom_style: atom_style to be used for the LAMMPS calculation (see LAMMPS docs).
            dimension: dimension to be used for the LAMMPS calculation (see LAMMPS docs).
            boundary: boundary to be used for the LAMMPS calculation (see LAMMPS docs).
            read_data: read_data to be used for the LAMMPS calculation (see LAMMPS docs).
            force_field: force field to be used for the LAMMPS calculation (see LAMMPS docs).
                Note that you should provide all the required information as a single string.
                In case of multiple lines expected in the input file,
                separate them with '\n' in force_field.
            keep_stages: If True, the string is formatted in a block structure with stage names
                and newlines that differentiate commands in the respective stages of the InputFile.
                If False, stage names are not printed and all commands appear in a single block.
        """
        if template is None:
            template = f"{TEMPLATE_DIR}/minimization.template"
        settings = {
            "units": units,
            "atom_style": atom_style,
            "dimension": dimension,
            "boundary": boundary,
            "read_data": read_data,
            "force_field": force_field,
        }

        super().__init__(
            template=template,
            settings=settings,
            calc_type="minimization",
            keep_stages=keep_stages,
        )

    @property
    def units(self) -> str:
        """The argument of the command 'units' passed to the generator."""
        return self.settings["units"]

    @property
    def atom_style(self) -> str:
        """The argument of the command 'atom_style' passed to the generator."""
        return self.settings["atom_style"]

    @property
    def dimension(self) -> int:
        """The argument of the command 'dimension' passed to the generator."""
        return self.settings["dimension"]

    @property
    def boundary(self) -> str:
        """The argument of the command 'boundary' passed to the generator."""
        return self.settings["boundary"]

    @property
    def read_data(self) -> str:
        """The argument of the command 'read_data' passed to the generator."""
        return self.settings["read_data"]

    @property
    def force_field(self) -> str:
        """The details of the force field commands passed to the generator."""
        return self.settings["force_field"]



"""This module implements an interface to enumlib, Gus Hart's excellent Fortran
code for enumerating derivative structures.

This module depends on a compiled enumlib with the executables enum.x and
makestr.x available in the path. Please download the library at
https://github.com/msg-byu/enumlib and follow the instructions in the README to
compile these two executables accordingly.

Alternatively, ``enum.x`` can be provided by Enumlib.jl
(https://github.com/glwhart/Enumlib.jl), a from-scratch Julia reimplementation of
enumlib by one of its authors. It is a drop-in replacement: the
``struct_enum.in`` / ``struct_enum.out`` file contract is unchanged and
``makeStr.py`` is reused, so only ``enum.x`` is swapped. This module's own test
suite passes against it.

If you use this module, please cite:

    - Gus L. W. Hart and Rodney W. Forcade, "Algorithm for generating derivative
    structures," Phys. Rev. B 77 224115 (26 June 2008)

    - Gus L. W. Hart and Rodney W. Forcade, "Generating derivative structures from
    multilattices: Application to hcp alloys," Phys. Rev. B 80 014120 (July 2009)

    - Gus L. W. Hart, Lance J. Nelson, and Rodney W. Forcade, "Generating
    derivative structures at a fixed concentration," Comp. Mat. Sci. 59
    101-107 (March 2012)

    - Wiley S. Morgan, Gus L. W. Hart, Rodney W. Forcade, "Generating derivative
    superstructures for systems with high configurational freedom," Comp. Mat.
    Sci. 136 144-149 (May 2017)
"""

from __future__ import annotations

import fractions
import functools
import itertools
import logging
import math
import re
import subprocess
import warnings
from glob import glob
from shutil import which
from typing import TYPE_CHECKING

import numpy as np
from monty.dev import requires
from monty.tempfile import ScratchDir

from pymatgen.core import DummySpecies, PeriodicSite, Structure
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

if TYPE_CHECKING:
    from typing import ClassVar

    from pymatgen.core.structure import IStructure

logger = logging.getLogger(__name__)

# Favor the use of the newer "enum.x" by Gus Hart over "multienum.x"
ENUM_CMD = which("enum.x") or which("multienum.x")
# Prefer makestr.x at present
MAKESTR_CMD = which("makestr.x") or which("makeStr.x") or which("makeStr.py")


@functools.cache
def _is_enumlib_jl(enum_cmd: str | None) -> bool:
    """Detect whether the resolved enum.x is the Julia Enumlib.jl engine.

    Enumlib.jl responds to ``enum.x --version`` by printing a line containing
    the token ``Enumlib.jl`` (e.g. ``enum.x (Enumlib.jl) 0.3.2``). The legacy
    Fortran enum.x does not support ``--version``. The probe is cached so the
    subprocess is spawned at most once per process per command path, and it
    never raises: any failure is treated as "not Enumlib.jl".

    Args:
        enum_cmd (str | None): Resolved path to the enum.x executable.

    Returns:
        bool: True iff the executable identifies itself as Enumlib.jl.
    """
    if not enum_cmd:
        return False
    try:
        proc = subprocess.run(
            [enum_cmd, "--version"],
            capture_output=True,
            text=True,
            timeout=10,
            check=False,
        )
        return "Enumlib.jl" in (proc.stdout or "") + (proc.stderr or "")
    except Exception:
        # Never let the version probe crash enumeration: any failure
        # (FileNotFoundError, subprocess.TimeoutExpired, OSError, ...) simply
        # means we cannot confirm the Julia engine.
        return False


def _warn_if_not_enumlib_jl(enum_cmd: str | None) -> None:
    """Advise switching to Enumlib.jl when the legacy Fortran enum.x is used.

    Purely advisory: this never raises or changes enumeration behavior. The
    underlying probe is cached, so the subprocess runs at most once per command
    path; how often the warning itself is shown is up to the active warning
    filters (Python's default filter shows it once per call site).

    Args:
        enum_cmd (str | None): Resolved path to the enum.x executable.
    """
    if enum_cmd and not _is_enumlib_jl(enum_cmd):
        warnings.warn(
            "You appear to be using the legacy Fortran 'enum.x'. Consider installing "
            "Enumlib.jl (https://github.com/glwhart/Enumlib.jl), a faster, "
            "better-maintained drop-in replacement for enumlib.",
            UserWarning,
            stacklevel=2,
        )


@requires(
    ENUM_CMD and MAKESTR_CMD,  # type: ignore[arg-type]
    "EnumlibAdaptor requires the executables 'enum.x' or 'multienum.x' "
    "and 'makestr.x' or 'makeStr.py' to be in the path. Please download the "
    "library at https://github.com/msg-byu/enumlib and follow the instructions "
    "in the README to compile these two executables accordingly.",
)
class EnumlibAdaptor:
    """An adaptor for enumlib.

    Attributes:
        structures (list[Structure]): all enumerated structures.
    """

    amount_tol: ClassVar[float] = 1e-5

    def __init__(
        self,
        structure: Structure | IStructure,
        min_cell_size: int = 1,
        max_cell_size: int = 1,
        symm_prec: float = 0.1,
        enum_precision_parameter: float = 0.001,
        refine_structure: bool = False,
        check_ordered_symmetry: bool = True,
        timeout: float | None = None,
    ) -> None:
        """Initialize the adapter with a structure and some parameters.

        Args:
            structure (Structure): An input structure.
            min_cell_size (int): The minimum cell size wanted. Defaults to 1.
            max_cell_size (int): The maximum cell size wanted. Defaults to 1.
            symm_prec (float): Symmetry precision. Defaults to 0.1.
            enum_precision_parameter (float): Finite precision parameter for
                enumlib. Default of 0.001 is usually ok, but you might need to
                tweak it for certain cells.
            refine_structure (bool): If you are starting from a structure that
                has been relaxed via some electronic structure code,
                it is usually much better to start with symmetry determination
                and then obtain a refined structure. The refined structure have
                cell parameters and atomic positions shifted to the expected
                symmetry positions, which makes it much less sensitive precision
                issues in enumlib. If you are already starting from an
                experimental cif, refinement should have already been done and
                it is not necessary. Defaults to False.
            check_ordered_symmetry (bool): Whether to check the symmetry of
                the ordered sites. If the symmetry of the ordered sites is
                lower, the lowest symmetry ordered sites is included in the
                enumeration. This is important if the ordered sites break
                symmetry in a way that is important getting possible
                structures. But sometimes including ordered sites
                slows down enumeration to the point that it cannot be
                completed. Switch to False in those cases. Defaults to True.
            timeout (float): If specified, will kill enumlib after specified
                time in minutes. This can be useful for gracefully handling
                enumerations in a high-throughput context, for some enumerations
                which will not terminate in a realistic length of time.
        """
        if refine_structure:
            finder = SpacegroupAnalyzer(structure, symm_prec)
            self.structure = finder.get_refined_structure()
        else:
            self.structure = structure  # type: ignore[assignment]

        self.min_cell_size = min_cell_size
        self.max_cell_size = max_cell_size
        self.symm_prec = symm_prec
        self.enum_precision_parameter = enum_precision_parameter
        self.check_ordered_symmetry = check_ordered_symmetry
        self.timeout = timeout

        # Recommend the faster, better-maintained Enumlib.jl when the legacy
        # Fortran enum.x is detected. Advisory only; the version probe behind
        # this is cached, so repeated instantiation costs no extra subprocesses.
        _warn_if_not_enumlib_jl(ENUM_CMD)

    def run(self) -> None:
        """Run the enumeration."""
        # Work in a temporary directory
        with ScratchDir(".") as tmp_dir:
            logger.debug(f"Temp dir : {tmp_dir}")
            # Generate input files
            self._gen_input_file()

            # Perform the actual enumeration
            num_structs = self._run_multienum()

            # Read in the enumeration output as structures.
            if num_structs > 0:
                self.structures = self._get_structures(num_structs)
            else:
                raise EnumError("Unable to enumerate structure.")

    def _gen_input_file(self) -> None:
        """Generate the necessary struct_enum.in file for enumlib. See enumlib
        documentation for details.
        """
        coord_format = "{:.6f} {:.6f} {:.6f}"
        # Use symmetry finder to get the symmetrically distinct sites
        fitter = SpacegroupAnalyzer(self.structure, self.symm_prec)
        symmetrized_structure = fitter.get_symmetrized_structure()
        logger.debug(
            f"Spacegroup {fitter.get_space_group_symbol()} ({fitter.get_space_group_number()}) "
            f"with {len(symmetrized_structure.equivalent_sites)} distinct sites"
        )

        # Enumlib doesn"t work when the number of species get too large. To
        # simplify matters, we generate the input file only with disordered sites
        # and exclude the ordered sites from the enumeration. The fact that
        # different disordered sites with the exact same species may belong to
        # different equivalent sites is dealt with by having determined the
        # spacegroup earlier and labelling the species differently.

        # `index_species` and `index_amounts` store mappings between the indices
        # used in the enum input file, and the actual species and amounts.
        index_species = []
        index_amounts = []

        # Store the ordered sites, which are not enumerated.
        ordered_sites = []
        disordered_sites = []
        coord_str: list[str] = []
        for sites in symmetrized_structure.equivalent_sites:
            if sites[0].is_ordered:
                ordered_sites.append(sites)
            else:
                _sp_label: list[int] = []
                species = dict(sites[0].species.items())
                if sum(species.values()) < 1 - EnumlibAdaptor.amount_tol:
                    # Let us first make add a dummy element for every single
                    # site whose total occupancies don't sum to 1.
                    species[DummySpecies("X")] = 1 - sum(species.values())
                for sp, amt in species.items():
                    if sp not in index_species:
                        index_species.append(sp)
                        _sp_label.append(len(index_species) - 1)
                        index_amounts.append(amt * len(sites))
                    else:
                        ind = index_species.index(sp)
                        _sp_label.append(ind)
                        index_amounts[ind] += amt * len(sites)
                sp_label: str = "/".join(f"{i}" for i in sorted(_sp_label))
                coord_str.extend(f"{coord_format.format(*site.coords)} {sp_label}" for site in sites)
                disordered_sites.append(sites)

        def get_sg_number(ss) -> int:
            finder = SpacegroupAnalyzer(Structure.from_sites(ss), self.symm_prec)
            return finder.get_space_group_number()

        target_sg_num = get_sg_number(list(symmetrized_structure))
        curr_sites = list(itertools.chain.from_iterable(disordered_sites))
        sg_num = get_sg_number(curr_sites)
        ordered_sites = sorted(ordered_sites, key=len)
        logger.debug(f"Disordered sites has sg # {sg_num}")
        self.ordered_sites = []

        # Progressively add ordered sites to our disordered sites
        # until we match the symmetry of our input structure
        if self.check_ordered_symmetry:
            while sg_num != target_sg_num and len(ordered_sites) > 0:
                sites = ordered_sites.pop(0)
                temp_sites = list(curr_sites) + sites
                new_sg_num = get_sg_number(temp_sites)
                if sg_num != new_sg_num:
                    logger.debug(f"Adding {sites[0].specie} in enum. New sg # {new_sg_num}")
                    index_species.append(sites[0].specie)
                    index_amounts.append(len(sites))
                    coord_str.extend(f"{coord_format.format(*site.coords)} {len(index_species) - 1}" for site in sites)
                    disordered_sites.append(sites)
                    curr_sites = temp_sites
                    sg_num = new_sg_num
                else:
                    self.ordered_sites.extend(sites)

        for sites in ordered_sites:
            self.ordered_sites.extend(sites)

        self.index_species = index_species

        lattice = self.structure.lattice

        output = [self.structure.formula, "bulk"]
        output.extend(coord_format.format(*vec) for vec in lattice.matrix)
        output.extend((f"{len(index_species)}", f"{len(coord_str)}"))
        output.extend(coord_str)

        output.extend(
            (
                f"{self.min_cell_size} {self.max_cell_size}",
                str(self.enum_precision_parameter),
                "full",
            )
        )

        n_disordered = sum(len(s) for s in disordered_sites)
        base = int(
            n_disordered
            * math.lcm(
                *(
                    fraction.limit_denominator(n_disordered * self.max_cell_size).denominator
                    for fraction in map(fractions.Fraction, index_amounts)
                )
            )
        )

        # This multiplicative factor of 10 is to prevent having too small bases
        # which can lead to rounding issues in the next step.
        # An old bug was that a base was set to 8, with a conc of 0.4:0.6. That
        # resulted in a range that overlaps and a conc of 0.5 satisfying this
        # enumeration. See Cu7Te5.cif test file.
        base *= 10

        # base = n_disordered # 10 ** math.ceil(math.log10(n_disordered))
        # To get a reasonable number of structures, we fix concentrations to the
        # range expected in the original structure.
        total_amounts = sum(index_amounts)
        for amt in index_amounts:
            conc = amt / total_amounts

            if abs(conc * base - round(conc * base)) < 1e-5:
                output.append(f"{round(conc * base)} {round(conc * base)} {base}")
            else:
                min_conc = math.floor(conc * base)
                output.append(f"{min_conc - 1} {min_conc + 1} {base}")
        output.append("")

        logger.debug("Generated input file:\n" + "\n".join(output))
        with open("struct_enum.in", mode="w", encoding="utf-8") as file:
            file.write("\n".join(output))

    def _run_multienum(self) -> int:
        """Run enumlib to get multiple structure.

        Returns:
            int: number of structures.
        """
        if ENUM_CMD is None:
            raise RuntimeError("enumlib is not available")

        with subprocess.Popen([ENUM_CMD], stdout=subprocess.PIPE, stdin=subprocess.PIPE, close_fds=True) as process:
            timeout = self.timeout * 60 if self.timeout is not None else None

            try:
                output = process.communicate(timeout=timeout)[0].decode("utf-8")
            except subprocess.TimeoutExpired as exc:
                process.kill()
                process.wait()
                raise TimeoutError(f"Enumeration took more than timeout {self.timeout} minutes") from exc

        count = 0
        start_count = False
        for line in output.strip().split("\n"):
            if line.strip().endswith("RunTot"):
                start_count = True
            elif start_count and re.match(r"\d+\s+.*", line.strip()):
                count = int(line.split()[-1])
        logger.debug(f"Enumeration resulted in {count} structures")
        return count

    def _get_structures(self, num_structs: int) -> list[Structure]:
        if MAKESTR_CMD is None:
            raise RuntimeError("makestr.x is not available")

        structs: list[Structure] = []

        if ".py" in MAKESTR_CMD:
            options: tuple[str, ...] = ("-input", "struct_enum.out", str(1), str(num_structs))
        else:
            options = ("struct_enum.out", str(0), str(num_structs - 1))

        with subprocess.Popen(
            [MAKESTR_CMD, *options],
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            close_fds=True,
        ) as rs:
            _stdout, stderr = rs.communicate()

        if stderr:
            logger.warning(stderr.decode())

        # Sites retrieved from enumlib will lack site properties
        # to ensure consistency, we keep track of what site properties
        # are missing and set them to None
        # TODO: improve this by mapping ordered structure to original
        # disordered structure, and retrieving correct site properties
        disordered_site_properties: dict = {}

        if len(self.ordered_sites) > 0:
            original_latt = self.ordered_sites[0].lattice
            # Need to strip sites of site_properties, which would otherwise
            # result in an index error. Hence Structure is reconstructed in
            # the next step.
            site_properties: dict = {}
            for site in self.ordered_sites:
                for k, v in site.properties.items():
                    disordered_site_properties[k] = None
                    if k in site_properties:
                        site_properties[k].append(v)
                    else:
                        site_properties[k] = [v]
            ordered_structure = Structure(
                original_latt,
                [site.species for site in self.ordered_sites],
                [site.frac_coords for site in self.ordered_sites],
                site_properties=site_properties,
            )
            inv_org_latt = np.linalg.inv(original_latt.matrix)
        else:
            ordered_structure = None  # type: ignore[assignment]
            inv_org_latt = None  # type: ignore[assignment]

        for file in glob("vasp.*"):
            with open(file, encoding="utf-8") as _file:
                data = _file.read()
                data = re.sub(r"scale factor", "1", data)
                data = re.sub(r"(\d+)-(\d+)", r"\1 -\2", data)
                poscar = Poscar.from_str(data, self.index_species)  # type: ignore[arg-type]
                sub_structure = poscar.structure
                # Enumeration may have resulted in a super lattice. We need to
                # find the mapping from the new lattice to the old lattice, and
                # perform supercell construction if necessary.
                new_latt = sub_structure.lattice

                sites = []

                if len(self.ordered_sites) > 0:
                    transformation = np.dot(new_latt.matrix, inv_org_latt)  # type:ignore[arg-type]
                    transformation = [[round(cell) for cell in row] for row in transformation]
                    logger.debug(f"Supercell matrix: {transformation}")
                    struct = ordered_structure * transformation
                    sites.extend([site.to_unit_cell() for site in struct])
                    super_latt = sites[-1].lattice
                else:
                    super_latt = new_latt

                for site in sub_structure:
                    if site.specie.symbol != "X":  # We exclude vacancies.
                        sites.append(
                            PeriodicSite(
                                site.species,
                                site.frac_coords,
                                super_latt,
                                to_unit_cell=True,
                                properties=disordered_site_properties,
                            )
                        )
                    else:
                        logger.debug("Skipping sites that include species X.")
                structs.append(Structure.from_sites(sorted(sites)))  # type:ignore[arg-type]

        logger.debug(f"Read in a total of {num_structs} structures.")
        return structs


class EnumError(BaseException):
    """Error subclass for enumeration errors."""



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib import colors
from pytest import approx

from pymatgen.core import Structure
from pymatgen.core import constants as const
from pymatgen.io.phonopy import get_gruneisen_ph_bs_symm_line, get_gruneisenparameter
from pymatgen.phonon.gruneisen import GruneisenParameter
from pymatgen.phonon.plotter import GruneisenPhononBandStructureSymmLine, GruneisenPhononBSPlotter, GruneisenPlotter
from pymatgen.util.testing import TEST_FILES_DIR, MatSciTest

try:
    import phonopy
    from phonopy.phonon.dos import TotalDos
except ImportError as exc:
    print(exc)
    phonopy = TotalDos = None

TEST_DIR = f"{TEST_FILES_DIR}/phonon/gruneisen"


class TestGruneisenPhononBandStructureSymmLine(MatSciTest):
    def setup_method(self) -> None:
        self.bs_symm_line = get_gruneisen_ph_bs_symm_line(
            gruneisen_path=f"{TEST_DIR}/gruneisen_eq_plus_minus_InP.yaml",
            structure_path=f"{TEST_DIR}/eq/POSCAR_InP",
            fit=True,
        )

    def test_plot(self):
        plotter = GruneisenPhononBSPlotter(bs=self.bs_symm_line)
        ax = plotter.get_plot_gs()
        assert isinstance(ax, plt.Axes)

    def test_ph_plot_w_gruneisen(self):
        plotter = GruneisenPhononBSPlotter(bs=self.bs_symm_line)
        ax = plotter.get_plot_gs(plot_ph_bs_with_gruneisen=True, units="THz", cmap=["red", "royalblue"])
        assert ax.get_ylabel() == "Frequencies (THz)"
        assert ax.get_xlabel() == "$\\mathrm{Wave\\ Vector}$"
        assert ax.get_figure()._localaxes[-1].get_ylabel() == "$\\gamma \\ \\mathrm{(logarithmized)}$"
        assert len(ax._children) == plotter.n_bands + 1  # check for number of bands
        # check for x and y data is really the band-structure data
        for inx, band in enumerate(plotter._bs.bands):
            xy_data = {
                "x": [point[0] for point in ax._children[inx].get_offsets().data],
                "y": [point[1] for point in ax._children[inx].get_offsets().data],
            }
            assert band == approx(xy_data["y"])
            assert plotter._bs.distance == approx(xy_data["x"])

        # check if color bar max value matches maximum gruneisen parameter value
        data = plotter.bs_plot_data()

        # get reference min and max Grüneisen parameter values
        max_gruneisen = np.array(data["gruneisen"]).max()
        min_gruneisen = np.array(data["gruneisen"]).min()

        norm = colors.SymLogNorm(
            vmin=min_gruneisen,
            vmax=max_gruneisen,
            linthresh=1e-2,
            linscale=1,
        )

        assert max(norm.inverse(ax.get_figure()._localaxes[-1].get_yticks())) == approx(max_gruneisen)
        assert isinstance(ax, plt.Axes)

    def test_as_dict_from_dict(self):
        new_dict = self.bs_symm_line.as_dict()
        self.new_bs_symm_line = GruneisenPhononBandStructureSymmLine.from_dict(new_dict)
        plotter = GruneisenPhononBSPlotter(bs=self.new_bs_symm_line)
        ax = plotter.get_plot_gs()
        assert isinstance(ax, plt.Axes)


class TestGruneisenParameterDebyeTempShim:
    def test_debye_temp_phonopy_run_debye_frequency_uses_new_signature(self, monkeypatch: pytest.MonkeyPatch):
        class FakeTotalDos:
            def __init__(self) -> None:
                self.freq_max_fit: float | None = None

            def run_debye_frequency(self, freq_max_fit=None) -> float:
                self.freq_max_fit = freq_max_fit
                return 2.5

        fake_tdos = FakeTotalDos()
        monkeypatch.setattr(GruneisenParameter, "tdos", property(lambda _self: fake_tdos))
        gruneisen = GruneisenParameter(
            qpoints=np.zeros((1, 3)),
            gruneisen=np.zeros((1, 1)),
            frequencies=np.zeros((1, 1)),
            structure=Structure(np.eye(3), ["H"], [[0, 0, 0]]),
        )

        debye_temp = gruneisen.debye_temp_phonopy(freq_max_fit=1.23)

        assert fake_tdos.freq_max_fit == approx(1.23)
        assert debye_temp == approx(
            const.value("Planck constant") * 2.5 * const.tera / const.value("Boltzmann constant")
        )

    def test_debye_temp_phonopy_run_debye_frequency_internal_failure(self, monkeypatch: pytest.MonkeyPatch):
        class FakeTotalDos:
            def run_debye_frequency(self, freq_max_fit=None) -> float:
                raise AttributeError("'TempMesh' object has no attribute 'primitive'")

        monkeypatch.setattr(GruneisenParameter, "tdos", property(lambda _self: FakeTotalDos()))
        gruneisen = GruneisenParameter(
            qpoints=np.zeros((1, 3)),
            gruneisen=np.zeros((1, 1)),
            frequencies=np.zeros((1, 1)),
            structure=Structure(np.eye(3), ["H"], [[0, 0, 0]]),
        )

        with pytest.raises(RuntimeError, match="cannot compute Debye frequency"):
            gruneisen.debye_temp_phonopy()


@pytest.mark.skipif(TotalDos is None, reason="Phonopy not present")
class TestGruneisenParameter(MatSciTest):
    def setup_method(self) -> None:
        self.gruneisen_obj = get_gruneisenparameter(
            f"{TEST_DIR}/gruneisen_mesh_InP.yaml",
            structure_path=f"{TEST_DIR}/eq/POSCAR_InP",
        )
        self.gruneisen_obj_small = get_gruneisenparameter(
            f"{TEST_DIR}/gruneisen_mesh_only_one_q_InP.yaml",
            structure_path=f"{TEST_DIR}/eq/POSCAR_InP",
        )
        self.gruneisen_obj_Si = get_gruneisenparameter(
            f"{TEST_DIR}/gruneisen_mesh_Si.yaml",
            structure_path=f"{TEST_DIR}/eq/POSCAR_Si",
        )

    def test_plot(self):
        plotter = GruneisenPlotter(self.gruneisen_obj)
        ax = plotter.get_plot(units="mev")
        assert isinstance(ax, plt.Axes)

    def test_as_from_dict(self):
        new_dict = self.gruneisen_obj.as_dict()
        self.gruneisen_obj2 = GruneisenParameter.from_dict(new_dict)

    def test_frequencies(self):
        assert self.gruneisen_obj_small.frequencies == approx(
            [0.12642146, 0.12642146, 0.25272004, 8.85202452, 8.85202452, 9.66016595]
        )

    def test_multi(self):
        assert self.gruneisen_obj_small.multiplicities[0] == 1
        assert self.gruneisen_obj.multiplicities[0] == 2

    def test_gruneisen(self):
        assert self.gruneisen_obj_small.gruneisen[0] == approx(-0.6176464482)
        assert self.gruneisen_obj_small.gruneisen[5] == approx(1.7574050911)

    def test_tdos(self):
        tot_dos = self.gruneisen_obj.tdos
        assert isinstance(tot_dos, phonopy.phonon.dos.TotalDos)

    def test_phdos(self):
        assert self.gruneisen_obj.phdos.cv(298.15) == approx(45.17772584681599)

    def test_average_gruneisen(self):
        assert self.gruneisen_obj.average_gruneisen() == approx(1.164231026696211)
        assert self.gruneisen_obj.average_gruneisen(squared=False) == approx(0.849759667411049)
        assert self.gruneisen_obj.average_gruneisen(limit_frequencies="debye") == approx(0.848865124114612)
        assert self.gruneisen_obj.average_gruneisen(limit_frequencies="acoustic") == approx(1.283180896570312)
        assert self.gruneisen_obj_Si.average_gruneisen() == approx(1.1090815951892143)

    def test_thermal_conductivity_slack(self):
        assert self.gruneisen_obj.thermal_conductivity_slack() == approx(77.97582174520458)
        assert self.gruneisen_obj.thermal_conductivity_slack(t=300) == approx(88.94562145031158)
        assert self.gruneisen_obj_Si.thermal_conductivity_slack(t=300) == approx(127.69008331982265)

    def test_debye_temp_phonopy(self):
        # This is the correct conversion when starting from THz in the debye_freq
        try:
            debye_temp = self.gruneisen_obj_small.debye_temp_phonopy()
        except RuntimeError as exc:
            pytest.skip(str(exc))

        assert debye_temp == approx(473.31932718764284)

    def test_acoustic_debye_temp(self):
        assert self.gruneisen_obj_small.acoustic_debye_temp == approx(317.54811309631845)
        assert self.gruneisen_obj.acoustic_debye_temp == approx(342.2046198151735)
        assert self.gruneisen_obj_Si.acoustic_debye_temp == approx(526.0725636300882)



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# distutils: language = c

from libc.stddef cimport ptrdiff_t


cdef const int BUFFER_SIZE = 1048576


cdef extern from *:
    """
    #define FFC_IMPL
    #include "ffc.h"

    static const ffc_parse_options PMG_FFC_PARSE_OPTIONS = {
        FFC_PRESET_GENERAL |
            FFC_FORMAT_FLAG_ALLOW_LEADING_PLUS |
            FFC_FORMAT_FLAG_SKIP_WHITE_SPACE,
        '.'
    };

    static inline int pmg_ffc_parse_one(
        const char* first,
        const char* last,
        double* value,
        const char** next
    ) {
        ffc_result result = ffc_from_chars_double_options(
            first,
            last,
            value,
            PMG_FFC_PARSE_OPTIONS
        );
        if (result.outcome != FFC_OUTCOME_OK) {
            return 0;
        }
        *next = result.ptr;
        return 1;
    }
    """
    int pmg_ffc_parse_one(
        const char* first,
        const char* last,
        double* value,
        const char** next,
    ) noexcept nogil


cdef const char* parse_doubles(
    const char* start,
    Py_ssize_t length,
    double* out,
    Py_ssize_t max_count,
    Py_ssize_t* parsed,
) noexcept nogil:
    """Parse doubles from a char buffer into ``out`` and return the stop pointer."""
    cdef const char* ptr = start
    cdef const char* end = start + length
    cdef const char* next_ptr
    cdef Py_ssize_t count = 0

    while count < max_count and ptr < end:
        if pmg_ffc_parse_one(ptr, end, &out[count], &next_ptr) == 0:
            break
        ptr = next_ptr
        count += 1

    parsed[0] = count
    return ptr


cdef Py_ssize_t last_whitespace(const unsigned char[::1] data) noexcept nogil:
    """Return the index after the last ASCII whitespace byte, or -1."""
    cdef Py_ssize_t idx = data.shape[0] - 1
    cdef unsigned char value

    while idx >= 0:
        value = data[idx]
        # ASCII whitespace: 9-13 (tab, LF, VT, FF, CR) and 32 (space).
        if (9 <= value <= 13) or value == 32:
            return idx + 1
        idx -= 1

    return -1


cdef bint only_whitespace_left(
    const unsigned char[::1] data, Py_ssize_t start, Py_ssize_t end,
) noexcept nogil:
    """Return whether ``data[start:end]`` contains only ASCII whitespace."""
    cdef Py_ssize_t idx = start
    cdef unsigned char value

    while idx < end:
        value = data[idx]
        if not ((9 <= value <= 13) or value == 32):
            return False
        idx += 1

    return True


def parse_n_doubles(file, double[::1] out, Py_ssize_t nelem=-1):
    """Parse doubles from a binary file object into ``out``.

    The file is read in 1 MiB chunks. On return, the file position follows the
    stop pointer returned by the parser without rewinding trailing whitespace.

    The ``file`` must be a seekable binary stream (the function uses
    ``tell()``/``seek()`` to leave the cursor right after the last parsed
    number). Backward seeks on compressed streams (e.g. ``gzip.GzipFile``,
    ``bz2.BZ2File``) are implemented as decompress-from-start and can be
    dramatically slower than on uncompressed files; prefer decompressing
    upstream when parsing large compressed inputs.

    Returns:
        int: Parsed element count.

    Raises:
        TypeError: If ``file`` is not a binary stream.
        ValueError: If ``nelem`` exceeds ``out.shape[0]``.
    """
    cdef Py_ssize_t limit = out.shape[0] if nelem < 0 else nelem
    if limit > out.shape[0]:
        raise ValueError("nelem exceeds output length")

    # Validate the file is binary up front, before mutating state via tell()/read().
    # A zero-byte read returns the empty bytes/str of the underlying mode without
    # advancing the file position.
    probe = file.read(0)
    if not isinstance(probe, bytes):
        raise TypeError("parse_n_doubles requires a binary file object")

    cdef Py_ssize_t total_parsed = 0
    cdef Py_ssize_t parsed = 0
    cdef Py_ssize_t remaining = limit
    cdef Py_ssize_t parse_len = 0
    cdef Py_ssize_t stop_offset
    cdef Py_ssize_t parsed_offset = 0
    cdef Py_ssize_t base_pos = file.tell()
    cdef Py_ssize_t consumed = 0
    cdef const char* start
    cdef const char* stop
    cdef bytes chunk
    cdef bytearray buffer = bytearray()
    cdef bint eof = False

    # Hold the memoryview only inside each iteration; ``bytearray.extend``/``del``
    # cannot resize the buffer while a memoryview is exported on it.
    cdef const unsigned char[::1] view
    cdef bint needs_more_data

    while remaining > 0:
        chunk = file.read(BUFFER_SIZE)

        if chunk:
            buffer.extend(chunk)
        else:
            eof = True

        if not buffer:
            break

        view = buffer
        parse_len = view.shape[0] if eof else last_whitespace(view)

        needs_more_data = False
        if parse_len < 0:
            if eof:
                parse_len = view.shape[0]
            else:
                needs_more_data = True
        elif parse_len == 0 and not eof:
            needs_more_data = True

        if needs_more_data:
            view = None
            continue

        start = <const char*>&view[0]
        stop = parse_doubles(start, parse_len, &out[total_parsed], remaining, &parsed)
        stop_offset = <ptrdiff_t>(stop - start)
        parsed_offset = stop_offset
        total_parsed += parsed
        remaining -= parsed

        if parsed == 0 or remaining == 0:
            view = None
            file.seek(base_pos + consumed + stop_offset)
            return total_parsed

        if stop_offset < parse_len:
            if not only_whitespace_left(view, stop_offset, parse_len):
                view = None
                file.seek(base_pos + consumed + stop_offset)
                return total_parsed

        consumed += parse_len
        view = None  # release before mutating the bytearray
        del buffer[:parse_len]

        if eof:
            break

    file.seek(base_pos + consumed + parsed_offset - parse_len)
    return total_parsed



"""Microbenchmarks for hot paths in pymatgen.core.

Run manually:
    uv run python tests/performance/bench_core.py

This script is intentionally standalone (no pytest, no xdist) to give
stable timings. Each benchmark is calibrated to run for ~0.2 s and reports
mean ns/call across multiple repeats.
"""

from __future__ import annotations

import timeit
import warnings
from typing import TYPE_CHECKING

import numpy as np

from pymatgen.core import Composition, Element, Lattice, Species
from pymatgen.core.periodic_table import get_el_sp
from pymatgen.optimization.neighbors import find_points_in_spheres
from pymatgen.util.coord_cython import is_coord_subset_pbc, pbc_shortest_vectors

if TYPE_CHECKING:
    from collections.abc import Callable

# Silence Element.X NaN warnings so they do not skew the X-accessing benches.
warnings.filterwarnings("ignore", message=r"No Pauling electronegativity.*")


def _time(stmt: Callable[[], object], *, target_s: float = 0.2, repeats: int = 5) -> tuple[float, float]:
    """Return (mean ns/call, stdev ns/call) over `repeats` runs.

    Calibrates `number` so each run lasts ~target_s.
    """
    # Calibration: find a `number` that takes >= ~10 ms.
    number = 1
    while True:
        t = timeit.timeit(stmt, number=number)
        if t >= 0.01:
            break
        number *= 10
    # Now scale to target_s.
    number = max(1, int(number * (target_s / max(t, 1e-9))))
    samples = [timeit.timeit(stmt, number=number) / number for _ in range(repeats)]
    mean = sum(samples) / len(samples)
    var = sum((s - mean) ** 2 for s in samples) / len(samples)
    stdev = var**0.5
    return mean * 1e9, stdev * 1e9


def _fmt(ns: float, sd: float) -> str:
    if ns >= 1e6:
        return f"{ns / 1e6:8.2f} ms  ± {sd / 1e6:.2f}"
    if ns >= 1e3:
        return f"{ns / 1e3:8.2f} µs  ± {sd / 1e3:.2f}"
    return f"{ns:8.1f} ns  ± {sd:.1f}"


def run(name: str, stmt: Callable[[], object]) -> None:
    ns, sd = _time(stmt)
    print(f"  {name:<48s} {_fmt(ns, sd)}")


def main() -> None:
    print("=" * 80)
    print("pymatgen.core microbenchmarks")
    print("=" * 80)

    # --- Composition parsing ------------------------------------------------
    print("\n[Composition] formula parsing")
    formulas_simple = ("LiFePO4", "Fe2O3", "NaCl", "H2O")
    formulas_complex = ("Li3Fe2(PO4)3", "Y3N@C80", "(NH4)2[PtCl6]", "Ca5(PO4)3F")

    def _parse_simple() -> None:
        for f in formulas_simple:
            Composition(f)

    def _parse_complex() -> None:
        for f in formulas_complex:
            Composition(f)

    run("Composition(simple) x4", _parse_simple)
    run("Composition(nested-parens) x4", _parse_complex)

    # Same formula repeated — measures benefit of caching.
    def _parse_repeat() -> None:
        for _ in range(8):
            Composition("LiFePO4")

    run("Composition('LiFePO4') x8 (repeat)", _parse_repeat)

    # --- get_el_sp ----------------------------------------------------------
    print("\n[get_el_sp] dispatch")
    run("get_el_sp('Fe')", lambda: get_el_sp("Fe"))
    run("get_el_sp('Fe2+')", lambda: get_el_sp("Fe2+"))
    run("get_el_sp(26)", lambda: get_el_sp(26))
    el_fe = Element("Fe")
    run("get_el_sp(Element('Fe'))", lambda: get_el_sp(el_fe))

    # --- Element properties -------------------------------------------------
    print("\n[Element] property access")
    fe = Element("Fe")
    he = Element("He")  # has no Pauling X

    run("Fe.X (200x)", lambda: [fe.X for _ in range(200)])
    run("He.X NaN path (200x)", lambda: [he.X for _ in range(200)])
    run("Fe.Z (200x)", lambda: [fe.Z for _ in range(200)])
    run("Fe.atomic_mass (200x)", lambda: [fe.atomic_mass for _ in range(200)])
    run("Fe.ionic_radii (50x)", lambda: [fe.ionic_radii for _ in range(50)])
    run("Fe.average_ionic_radius (50x)", lambda: [fe.average_ionic_radius for _ in range(50)])
    run("Fe.atomic_orbitals_eV (50x)", lambda: [fe.atomic_orbitals_eV for _ in range(50)])
    run("Fe.full_electronic_structure (50x)", lambda: [fe.full_electronic_structure for _ in range(50)])
    run("Fe.row (200x)", lambda: [fe.row for _ in range(200)])
    run("Fe.group (200x)", lambda: [fe.group for _ in range(200)])
    run("Fe.iupac_ordering (200x)", lambda: [fe.iupac_ordering for _ in range(200)])
    run("Fe.thermal_conductivity __getattr__ (200x)", lambda: [fe.thermal_conductivity for _ in range(200)])

    # --- Composition properties --------------------------------------------
    print("\n[Composition] derived properties")
    comp = Composition("Li3Fe2(PO4)3")
    big = Composition({Element(s): 1 for s in ("Li", "Na", "K", "Mg", "Ca", "Sr", "Ba", "Fe", "Co", "Ni", "O", "S")})

    run("comp.average_electroneg (200x)", lambda: [comp.average_electroneg for _ in range(200)])
    run("comp.total_electrons (200x)", lambda: [comp.total_electrons for _ in range(200)])
    run("comp.num_atoms (200x)", lambda: [comp.num_atoms for _ in range(200)])
    run("comp.weight (200x)", lambda: [comp.weight for _ in range(200)])
    run("comp.formula (200x)", lambda: [comp.formula for _ in range(200)])
    run("comp.reduced_formula (200x)", lambda: [comp.reduced_formula for _ in range(200)])
    run("big.average_electroneg (200x)", lambda: [big.average_electroneg for _ in range(200)])

    # --- Sorting compositions (uses Element.X and Composition.average_electroneg)
    print("\n[Composition] sorting (Element.X heavy)")
    comps = [Composition(f) for f in ("LiFePO4", "NaCl", "Fe2O3", "Al2O3", "MgO", "SiO2", "TiO2", "CaCO3")]
    run("sorted(comps) x10", lambda: [sorted(comps) for _ in range(10)])

    # --- Species creation ---------------------------------------------------
    print("\n[Species]")
    run("Species('Fe', 2)", lambda: Species("Fe", 2))
    run("Species.from_str('Fe2+')", lambda: Species.from_str("Fe2+"))

    # --- coord_cython hot paths --------------------------------------------
    print("\n[coord_cython] pbc_shortest_vectors / is_coord_subset_pbc")
    rng = np.random.default_rng(0)
    lattice = Lattice.cubic(10.0)

    # Small-ish (typical StructureMatcher inner call): 32 x 32.
    fc32a = rng.random((32, 3))
    fc32b = rng.random((32, 3))
    run("pbc_shortest_vectors 32x32", lambda: pbc_shortest_vectors(lattice, fc32a, fc32b))

    # Medium: 128 x 128.
    fc128a = rng.random((128, 3))
    fc128b = rng.random((128, 3))
    run("pbc_shortest_vectors 128x128", lambda: pbc_shortest_vectors(lattice, fc128a, fc128b))

    # Large: 256 x 256 (dominated by the i*j*27 inner kernel).
    fc256a = rng.random((256, 3))
    fc256b = rng.random((256, 3))
    run("pbc_shortest_vectors 256x256", lambda: pbc_shortest_vectors(lattice, fc256a, fc256b))

    # With mask (typical sparsity ~50%): also exercises the mask early-out.
    mask128 = (rng.random((128, 128)) > 0.5).astype(np.int64)
    run("pbc_shortest_vectors 128x128 +mask50", lambda: pbc_shortest_vectors(lattice, fc128a, fc128b, mask=mask128))

    # With return_d2 (used by StructureMatcher).
    run(
        "pbc_shortest_vectors 128x128 return_d2",
        lambda: pbc_shortest_vectors(lattice, fc128a, fc128b, return_d2=True),
    )

    # is_coord_subset_pbc: subset (16) inside superset (128).
    fc_sub = fc128a[:16]
    atol = np.array([1e-3, 1e-3, 1e-3])
    sub_mask = np.zeros((16, 128), dtype=np.int64)
    run(
        "is_coord_subset_pbc 16-in-128",
        lambda: is_coord_subset_pbc(fc_sub, fc128a, atol, sub_mask, pbc=(True, True, True)),
    )

    # --- neighbors.find_points_in_spheres ----------------------------------
    print("\n[neighbors] find_points_in_spheres")
    lat_mat = lattice.matrix
    pbc_all = np.array([1, 1, 1], dtype=np.int64)

    # Small: 32 atoms, 32 centers, cutoff 4 Å (typical short-range query).
    cc32 = (rng.random((32, 3)) * 10.0).astype(np.float64)
    ac32 = (rng.random((32, 3)) * 10.0).astype(np.float64)
    run(
        "find_points_in_spheres 32x32 r=4",
        lambda: find_points_in_spheres(ac32, cc32, 4.0, pbc_all, lat_mat),
    )

    # Medium: 128 atoms, 128 centers, cutoff 5 Å.
    cc128 = (rng.random((128, 3)) * 15.0).astype(np.float64)
    ac128 = (rng.random((128, 3)) * 15.0).astype(np.float64)
    lattice_med = Lattice.cubic(15.0).matrix
    run(
        "find_points_in_spheres 128x128 r=5",
        lambda: find_points_in_spheres(ac128, cc128, 5.0, pbc_all, lattice_med),
    )

    # Large: 256 atoms, 256 centers, larger cutoff.
    cc256 = (rng.random((256, 3)) * 20.0).astype(np.float64)
    ac256 = (rng.random((256, 3)) * 20.0).astype(np.float64)
    lattice_lg = Lattice.cubic(20.0).matrix
    run(
        "find_points_in_spheres 256x256 r=6",
        lambda: find_points_in_spheres(ac256, cc256, 6.0, pbc_all, lattice_lg),
    )

    # Larger cutoff with same medium box (more periodic images visited).
    run(
        "find_points_in_spheres 128x128 r=8",
        lambda: find_points_in_spheres(ac128, cc128, 8.0, pbc_all, lattice_med),
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

from io import BytesIO, StringIO

import numpy as np
import pytest

from pymatgen.optimization.fast_parser import parse_n_doubles


def assert_exact(actual: np.ndarray, expected: np.ndarray) -> None:
    assert actual.dtype == np.float64
    assert expected.dtype == np.float64
    assert np.array_equal(actual.view(np.uint64), expected.view(np.uint64))


def test_parse_n_doubles_reads_binary_file() -> None:
    data = b"header 1.5 2.5 3.5 tail"
    file = BytesIO(data)
    file.seek(len(b"header "))
    out = np.empty(3, dtype=np.float64)

    parsed = parse_n_doubles(file, out)

    assert parsed == 3
    assert file.tell() == len(b"header 1.5 2.5 3.5")
    assert file.read(1) == b" "
    assert_exact(out, np.array([1.5, 2.5, 3.5], dtype=np.float64))


def test_parse_n_doubles_respects_nelem() -> None:
    data = b"1 2 3 4"
    file = BytesIO(data)
    out = np.empty(4, dtype=np.float64)

    parsed = parse_n_doubles(file, out, nelem=2)

    assert parsed == 2
    assert file.tell() == len(b"1 2")
    assert file.read(1) == b" "
    assert_exact(out[:parsed], np.array([1.0, 2.0], dtype=np.float64))


def test_parse_n_doubles_rejects_text_file() -> None:
    file = StringIO("1 2 3")
    out = np.empty(3, dtype=np.float64)

    with pytest.raises(TypeError, match="binary file"):
        parse_n_doubles(file, out)

    # The probe is a zero-byte read, so the cursor must not move.
    assert file.tell() == 0


def test_parse_n_doubles_handles_values_larger_than_buffer() -> None:
    data = b" ".join(b"1.25" for _ in range(300_000)) + b" label"
    file = BytesIO(data)
    out = np.empty(300_000, dtype=np.float64)

    parsed = parse_n_doubles(file, out)

    assert parsed == 300_000
    assert file.tell() == len(data) - len(b" label")
    assert file.read(1) == b" "
    assert np.all(out == 1.25)


def test_parse_n_doubles_rejects_oversized_nelem() -> None:
    file = BytesIO(b"1 2 3")
    out = np.empty(2, dtype=np.float64)

    with pytest.raises(ValueError, match="nelem exceeds output length"):
        parse_n_doubles(file, out, nelem=3)


def test_parse_n_doubles_short_file_returns_partial_count() -> None:
    """EOF reached before ``nelem`` values: return partial count without raising."""
    data = b"1.0 2.0 3.0"
    file = BytesIO(data)
    out = np.empty(5, dtype=np.float64)

    parsed = parse_n_doubles(file, out, nelem=5)

    assert parsed == 3
    # Cursor should land right after the last successfully parsed number.
    assert file.tell() == len(data)
    assert_exact(out[:parsed], np.array([1.0, 2.0, 3.0], dtype=np.float64))


def test_parse_n_doubles_buffer_ends_at_newline() -> None:
    """A buffer terminating exactly on a whitespace byte is fully consumable."""
    data = b"1.0 2.0 3.0\n"
    file = BytesIO(data)
    out = np.empty(3, dtype=np.float64)

    parsed = parse_n_doubles(file, out)

    assert parsed == 3
    assert_exact(out, np.array([1.0, 2.0, 3.0], dtype=np.float64))


def test_parse_n_doubles_stops_on_invalid_token() -> None:
    """An invalid token before ``nelem`` is reached aborts parsing without consuming it."""
    data = b"1.0 2.0 not_a_number 3.0"
    file = BytesIO(data)
    out = np.empty(4, dtype=np.float64)

    parsed = parse_n_doubles(file, out, nelem=4)

    assert parsed == 2
    assert_exact(out[:parsed], np.array([1.0, 2.0], dtype=np.float64))
    # Cursor should sit at (or before) the invalid token so the caller can recover.
    remaining = file.read()
    assert b"not_a_number" in remaining


def test_parse_n_doubles_token_longer_than_buffer() -> None:
    """A single huge token (no embedded whitespace) is still parsed correctly."""
    # Build a token like "1." + many zeros + "5" that exceeds the 1 MiB buffer.
    huge_token = b"1." + b"0" * (1_500_000) + b"5"
    data = huge_token + b" 2.5"
    file = BytesIO(data)
    out = np.empty(2, dtype=np.float64)

    parsed = parse_n_doubles(file, out)

    assert parsed == 2
    # The huge token's value is mathematically 1.0...05 == 1.0 at double precision.
    assert out[0] == pytest.approx(1.0, abs=1e-300)
    assert out[1] == 2.5



from __future__ import annotations

import filecmp
import re

import pandas as pd
import pytest

from pymatgen.io.lammps.data import LammpsData
from pymatgen.io.lammps.inputs import LammpsInputFile, LammpsTemplateGen, write_lammps_inputs
from pymatgen.util.testing import TEST_FILES_DIR, MatSciTest

TEST_DIR = f"{TEST_FILES_DIR}/io/lammps"


class TestLammpsInputFile(MatSciTest):
    @classmethod
    def setup_class(cls):
        cls.filename = f"{TEST_DIR}/lgps.in"

    def test_from_file(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        assert lmp_input.ncomments == 3
        assert lmp_input.nstages == 1
        assert lmp_input.stages_names == ["Stage 1"]
        assert lmp_input.stages[0]["commands"][0][1] == "metal"
        assert lmp_input.stages == [
            {
                "stage_name": "Stage 1",
                "commands": [
                    ("units", "metal"),
                    ("atom_style", "full"),
                    ("dimension", "3"),
                    ("pair_style", "hybrid/overlay morse 15 coul/long 15"),  # codespell:ignore coul
                    ("kspace_style", "ewald 1e-4"),
                    ("boundary", "p p p"),
                    ("#", "2) System definition"),
                    ("read_data", "run_init.data"),
                    ("set", "type 1 charge 0.8803"),
                    ("set", "type 2 charge 1.2570"),
                    ("set", "type 3 charge 1.2580"),
                    ("set", "type 4 charge -1.048"),
                    ("neigh_modify", "every 1 delay 5 check yes"),
                    ("#", "3) Simulation settings"),
                    ("pair_coeff", "1 1 morse 0.0580 3.987 3.404"),
                    ("pair_coeff", "1 4 morse 0.0408 1.399 3.204"),
                    ("pair_coeff", "2 4 morse 0.3147 2.257 2.409"),
                    ("pair_coeff", "3 4 morse 0.4104 2.329 2.200"),
                    ("pair_coeff", "4 4 morse 0.0241 1.359 4.284"),
                    ("pair_coeff", "* * coul/long"),  # codespell:ignore coul
                    ("#", "Part A : energy minimization"),
                    ("thermo", "1"),
                    ("thermo_style", "custom step lx ly lz press pxx pyy pzz pe"),
                    ("dump", "dmp all atom 5 run.dump"),
                    ("min_style", "cg"),
                    ("fix", "1 all box/relax iso 0.0 vmax 0.001"),
                    ("minimize", "1.0e-16 1.0e-16 5000 10000"),
                    ("write_data", "run.data"),
                ],
            }
        ]

    def test_get_str(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        string = lmp_input.get_str()
        assert "# LAMMPS input generated from LammpsInputFile with pymatgen v" in string
        assert "\nunits metal\natom_style full\ndimension 3\npair_style hybrid/overlay morse 15 coul/long" in string
        assert "15\nkspace_style ewald 1e-4\nboundary p p p\n# 2) System definition" in string
        assert "\nread_data run_init.data\nset type 1 charge 0.8803\nset type 2 charge" in string
        assert "1.2570\nset type 3 charge 1.2580\nset type 4 charge -1.048\nneigh_modify" in string
        assert "every 1 delay 5 check yes\n# 3) Simulation settings\npair_coeff 1 1" in string
        assert "morse 0.0580 3.987 3.404\npair_coeff 1 4 morse 0.0408 1.399 3.204\npair_coeff" in string
        assert "2 4 morse 0.3147 2.257 2.409\npair_coeff 3 4 morse 0.4104 2.329 2.200\npair_coeff" in string
        assert "4 4 morse 0.0241 1.359 4.284\npair_coeff * * coul/long\n# Part A : energy" in string
        assert "minimization\nthermo 1\nthermo_style custom step lx ly lz press pxx pyy pzz" in string
        assert "pe\ndump dmp all atom 5 run.dump\nmin_style cg\nfix 1 all" in string
        assert "box/relax iso 0.0 vmax 0.001\nminimize 1.0e-16 1.0e-16 5000 10000\nwrite_data run.data" in string

    def test_from_str(self):
        string = """# LGPS

# 1) Initialization
units metal
atom_style full
dimension 3
pair_style hybrid/overlay &
morse 15 coul/long 15
kspace_style ewald 1e-4
boundary p p p

# 2) System definition
read_data run_init.data
set type 1 charge  0.8803
set type 2 charge  1.2570
set type 3 charge  1.2580
set type 4 charge -1.048
neigh_modify every 1 delay 5 check yes

# 3) Simulation settings
pair_coeff 1 1 morse 0.0580 3.987 3.404
pair_coeff 1 4 morse 0.0408 1.399 3.204
pair_coeff 2 4 morse 0.3147 2.257 2.409
pair_coeff 3 4 morse 0.4104 2.329 2.200
pair_coeff 4 4 morse 0.0241 1.359 4.284
pair_coeff * * coul/long

# Part A : energy minimization
thermo 1
thermo_style custom step lx ly lz press pxx pyy pzz pe
dump dmp all atom 5 run.dump

min_style cg
fix 1 all box/relax iso 0.0 vmax 0.001
minimize 1.0e-16 1.0e-16 5000 10000
write_data run.data"""

        lmp_input = LammpsInputFile().from_str(string)
        assert lmp_input.stages == LammpsInputFile().from_file(self.filename).stages

    def test_stages_names(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        assert lmp_input.stages_names == ["Stage 1"]

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        assert lmp_input.stages_names == [
            "Comment 1",
            "1) Initialization",
            "2) System definition",
            "3) Simulation settings",
            "Part A : energy minimization",
            "Stage 6",
        ]

        lmp_input.stages.append({"stage_name": "New stage", "commands": [("units", "metal")]})
        assert lmp_input.stages_names == [
            "Comment 1",
            "1) Initialization",
            "2) System definition",
            "3) Simulation settings",
            "Part A : energy minimization",
            "Stage 6",
            "New stage",
        ]

    def test_nstages(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        assert lmp_input.nstages == 1

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        assert lmp_input.nstages == 6

    def test_ncomments(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        assert lmp_input.ncomments == 3

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        assert lmp_input.ncomments == 1

    def test_get_args(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        units = lmp_input.get_args("units")
        assert units == "metal"

        sets = lmp_input.get_args("set")
        assert sets == [
            "type 1 charge 0.8803",
            "type 2 charge 1.2570",
            "type 3 charge 1.2580",
            "type 4 charge -1.048",
        ]

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        set1 = lmp_input.get_args("set", stage_name="2) System definition")
        assert set1[1] == "type 2 charge 1.2570"

    def test_contains_command(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        assert lmp_input.contains_command("set")
        assert not lmp_input.contains_command("sett")

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        assert lmp_input.contains_command("units")
        assert lmp_input.contains_command("units", stage_name="1) Initialization")
        assert not lmp_input.contains_command("units", stage_name="Stage 6")

    def test_set_args(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        lmp_input.set_args(command="units", argument="atomic")
        assert lmp_input.get_args("units") == "atomic"

        lmp_input2 = lmp_input
        lmp_input2.set_args(command="set", argument="new set 2", how=2, stage_name="Stage 1")
        assert lmp_input.stages == lmp_input2.stages
        lmp_input2.set_args(command="set", argument="new set", how="first")
        assert lmp_input.get_args("set")[0] == "new set"
        assert lmp_input.get_args("set")[1] == "type 2 charge 1.2570"

        lmp_input3 = LammpsInputFile().from_file(self.filename, keep_stages=True)
        lmp_input4 = LammpsInputFile().from_file(self.filename, keep_stages=True)
        lmp_input4.set_args(command="set", argument="new set 2", how=2, stage_name="Stage 1")
        assert lmp_input3.stages == lmp_input4.stages
        lmp_input4.set_args(
            command="set",
            argument="new set 2",
            how="first",
            stage_name="2) System definition",
        )
        assert lmp_input4.get_args("set", stage_name="2) System definition")[0] == "new set 2"
        assert lmp_input4.get_args("set", stage_name="2) System definition")[1] == "type 2 charge 1.2570"

    def test_add_stage(self):
        lmp_input = LammpsInputFile()
        lmp_input.add_stage(commands="units metal")
        assert lmp_input.stages == [{"stage_name": "Stage 1", "commands": [("units", "metal")]}]
        lmp_input.add_stage(commands={"pair_style": "eam"}, stage_name="Pair style")
        assert lmp_input.stages == [
            {"stage_name": "Stage 1", "commands": [("units", "metal")]},
            {"stage_name": "Pair style", "commands": [("pair_style", "eam")]},
        ]
        lmp_input.add_stage(commands=["boundary p p p", "atom_style full"], stage_name="Cell")
        assert lmp_input.stages == [
            {"stage_name": "Stage 1", "commands": [("units", "metal")]},
            {"stage_name": "Pair style", "commands": [("pair_style", "eam")]},
            {
                "stage_name": "Cell",
                "commands": [("boundary", "p p p"), ("atom_style", "full")],
            },
        ]

        lmp_input.add_stage(commands=["set type 2 charge 0.0"], after_stage="Stage 1")
        assert lmp_input.stages == [
            {"stage_name": "Stage 1", "commands": [("units", "metal")]},
            {"stage_name": "Stage 4", "commands": [("set", "type 2 charge 0.0")]},
            {"stage_name": "Pair style", "commands": [("pair_style", "eam")]},
            {
                "stage_name": "Cell",
                "commands": [("boundary", "p p p"), ("atom_style", "full")],
            },
        ]
        lmp_input.add_stage(stage_name="Empty stage")
        assert lmp_input.stages == [
            {"stage_name": "Stage 1", "commands": [("units", "metal")]},
            {"stage_name": "Stage 4", "commands": [("set", "type 2 charge 0.0")]},
            {"stage_name": "Pair style", "commands": [("pair_style", "eam")]},
            {
                "stage_name": "Cell",
                "commands": [("boundary", "p p p"), ("atom_style", "full")],
            },
            {"stage_name": "Empty stage", "commands": []},
        ]
        lmp_input.add_stage(
            {
                "stage_name": "New stage",
                "commands": [
                    ("newcommand", "newargs 1 2 3"),
                    ("newcommand2", "newargs 4 5 6"),
                ],
            }
        )
        assert lmp_input.stages == [
            {"stage_name": "Stage 1", "commands": [("units", "metal")]},
            {"stage_name": "Stage 4", "commands": [("set", "type 2 charge 0.0")]},
            {"stage_name": "Pair style", "commands": [("pair_style", "eam")]},
            {
                "stage_name": "Cell",
                "commands": [("boundary", "p p p"), ("atom_style", "full")],
            },
            {"stage_name": "Empty stage", "commands": []},
            {
                "stage_name": "New stage",
                "commands": [
                    ("newcommand", "newargs 1 2 3"),
                    ("newcommand2", "newargs 4 5 6"),
                ],
            },
        ]
        lmp_input.add_stage(
            {
                "stage_name": "New stage 2",
                "commands": [
                    ("newcommand", "newargs 1 2 3"),
                    ("newcommand2", "newargs 4 5 6"),
                ],
            },
            after_stage="Stage 4",
        )
        assert lmp_input.stages == [
            {"stage_name": "Stage 1", "commands": [("units", "metal")]},
            {"stage_name": "Stage 4", "commands": [("set", "type 2 charge 0.0")]},
            {
                "stage_name": "New stage 2",
                "commands": [
                    ("newcommand", "newargs 1 2 3"),
                    ("newcommand2", "newargs 4 5 6"),
                ],
            },
            {"stage_name": "Pair style", "commands": [("pair_style", "eam")]},
            {
                "stage_name": "Cell",
                "commands": [("boundary", "p p p"), ("atom_style", "full")],
            },
            {"stage_name": "Empty stage", "commands": []},
            {
                "stage_name": "New stage",
                "commands": [
                    ("newcommand", "newargs 1 2 3"),
                    ("newcommand2", "newargs 4 5 6"),
                ],
            },
        ]

    def test_merge_stages(self):
        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        lmp_input.merge_stages(["1) Initialization", "3) Simulation settings"])
        assert lmp_input.stages_names == [
            "Comment 1",
            "Merge of: 1) Initialization, 3) Simulation settings",
            "2) System definition",
            "Part A : energy minimization",
            "Stage 6",
        ]
        assert lmp_input.stages == [
            {"stage_name": "Comment 1", "commands": [("#", "LGPS")]},
            {
                "stage_name": "Merge of: 1) Initialization, 3) Simulation settings",
                "commands": [
                    ("units", "metal"),
                    ("atom_style", "full"),
                    ("dimension", "3"),
                    ("pair_style", "hybrid/overlay morse 15 coul/long 15"),
                    ("kspace_style", "ewald 1e-4"),
                    ("boundary", "p p p"),
                    ("pair_coeff", "1 1 morse 0.0580 3.987 3.404"),
                    ("pair_coeff", "1 4 morse 0.0408 1.399 3.204"),
                    ("pair_coeff", "2 4 morse 0.3147 2.257 2.409"),
                    ("pair_coeff", "3 4 morse 0.4104 2.329 2.200"),
                    ("pair_coeff", "4 4 morse 0.0241 1.359 4.284"),
                    ("pair_coeff", "* * coul/long"),
                ],
            },
            {
                "stage_name": "2) System definition",
                "commands": [
                    ("read_data", "run_init.data"),
                    ("set", "type 1 charge 0.8803"),
                    ("set", "type 2 charge 1.2570"),
                    ("set", "type 3 charge 1.2580"),
                    ("set", "type 4 charge -1.048"),
                    ("neigh_modify", "every 1 delay 5 check yes"),
                ],
            },
            {
                "stage_name": "Part A : energy minimization",
                "commands": [
                    ("thermo", "1"),
                    ("thermo_style", "custom step lx ly lz press pxx pyy pzz pe"),
                    ("dump", "dmp all atom 5 run.dump"),
                ],
            },
            {
                "stage_name": "Stage 6",
                "commands": [
                    ("min_style", "cg"),
                    ("fix", "1 all box/relax iso 0.0 vmax 0.001"),
                    ("minimize", "1.0e-16 1.0e-16 5000 10000"),
                    ("write_data", "run.data"),
                ],
            },
        ]

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        lmp_input.merge_stages(lmp_input.stages_names)
        assert len(lmp_input.stages_names) == 1
        assert lmp_input.stages == [
            {
                "stage_name": "Merge of: Comment 1, 1) Initialization, 2) System definition, 3) Simulation settings, "
                "Part A : "
                "energy minimization, Stage 6",
                "commands": [
                    ("#", "LGPS"),
                    ("units", "metal"),
                    ("atom_style", "full"),
                    ("dimension", "3"),
                    ("pair_style", "hybrid/overlay morse 15 coul/long 15"),
                    ("kspace_style", "ewald 1e-4"),
                    ("boundary", "p p p"),
                    ("read_data", "run_init.data"),
                    ("set", "type 1 charge 0.8803"),
                    ("set", "type 2 charge 1.2570"),
                    ("set", "type 3 charge 1.2580"),
                    ("set", "type 4 charge -1.048"),
                    ("neigh_modify", "every 1 delay 5 check yes"),
                    ("pair_coeff", "1 1 morse 0.0580 3.987 3.404"),
                    ("pair_coeff", "1 4 morse 0.0408 1.399 3.204"),
                    ("pair_coeff", "2 4 morse 0.3147 2.257 2.409"),
                    ("pair_coeff", "3 4 morse 0.4104 2.329 2.200"),
                    ("pair_coeff", "4 4 morse 0.0241 1.359 4.284"),
                    ("pair_coeff", "* * coul/long"),
                    ("thermo", "1"),
                    ("thermo_style", "custom step lx ly lz press pxx pyy pzz pe"),
                    ("dump", "dmp all atom 5 run.dump"),
                    ("min_style", "cg"),
                    ("fix", "1 all box/relax iso 0.0 vmax 0.001"),
                    ("minimize", "1.0e-16 1.0e-16 5000 10000"),
                    ("write_data", "run.data"),
                ],
            }
        ]

    def test_remove_stage(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        lmp_input.remove_stage("Stage 1")
        assert lmp_input.stages == []

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        lmp_input.remove_stage("1) Initialization")
        assert lmp_input.stages_names == [
            "Comment 1",
            "2) System definition",
            "3) Simulation settings",
            "Part A : energy minimization",
            "Stage 6",
        ]

    def test_rename_stage(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        lmp_input.rename_stage("Stage 1", "Global stage")
        assert lmp_input.stages_names == ["Global stage"]

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        lmp_input.rename_stage("Stage 6", "Final stage")
        assert lmp_input.stages_names == [
            "Comment 1",
            "1) Initialization",
            "2) System definition",
            "3) Simulation settings",
            "Part A : energy minimization",
            "Final stage",
        ]

    def test_remove_command(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        lmp_input.remove_command("set")
        assert lmp_input.stages == [
            {
                "stage_name": "Stage 1",
                "commands": [
                    ("units", "metal"),
                    ("atom_style", "full"),
                    ("dimension", "3"),
                    ("pair_style", "hybrid/overlay morse 15 coul/long 15"),
                    ("kspace_style", "ewald 1e-4"),
                    ("boundary", "p p p"),
                    ("#", "2) System definition"),
                    ("read_data", "run_init.data"),
                    ("neigh_modify", "every 1 delay 5 check yes"),
                    ("#", "3) Simulation settings"),
                    ("pair_coeff", "1 1 morse 0.0580 3.987 3.404"),
                    ("pair_coeff", "1 4 morse 0.0408 1.399 3.204"),
                    ("pair_coeff", "2 4 morse 0.3147 2.257 2.409"),
                    ("pair_coeff", "3 4 morse 0.4104 2.329 2.200"),
                    ("pair_coeff", "4 4 morse 0.0241 1.359 4.284"),
                    ("pair_coeff", "* * coul/long"),
                    ("#", "Part A : energy minimization"),
                    ("thermo", "1"),
                    ("thermo_style", "custom step lx ly lz press pxx pyy pzz pe"),
                    ("dump", "dmp all atom 5 run.dump"),
                    ("min_style", "cg"),
                    ("fix", "1 all box/relax iso 0.0 vmax 0.001"),
                    ("minimize", "1.0e-16 1.0e-16 5000 10000"),
                    ("write_data", "run.data"),
                ],
            }
        ]

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        lmp_input._add_command(command="set type 1 charge 0.0", stage_name="1) Initialization")
        lmp_input.remove_command("set")
        assert lmp_input.stages == [
            {"stage_name": "Comment 1", "commands": [("#", "LGPS")]},
            {
                "stage_name": "1) Initialization",
                "commands": [
                    ("units", "metal"),
                    ("atom_style", "full"),
                    ("dimension", "3"),
                    ("pair_style", "hybrid/overlay morse 15 coul/long 15"),
                    ("kspace_style", "ewald 1e-4"),
                    ("boundary", "p p p"),
                ],
            },
            {
                "stage_name": "2) System definition",
                "commands": [
                    ("read_data", "run_init.data"),
                    ("neigh_modify", "every 1 delay 5 check yes"),
                ],
            },
            {
                "stage_name": "3) Simulation settings",
                "commands": [
                    ("pair_coeff", "1 1 morse 0.0580 3.987 3.404"),
                    ("pair_coeff", "1 4 morse 0.0408 1.399 3.204"),
                    ("pair_coeff", "2 4 morse 0.3147 2.257 2.409"),
                    ("pair_coeff", "3 4 morse 0.4104 2.329 2.200"),
                    ("pair_coeff", "4 4 morse 0.0241 1.359 4.284"),
                    ("pair_coeff", "* * coul/long"),
                ],
            },
            {
                "stage_name": "Part A : energy minimization",
                "commands": [
                    ("thermo", "1"),
                    ("thermo_style", "custom step lx ly lz press pxx pyy pzz pe"),
                    ("dump", "dmp all atom 5 run.dump"),
                ],
            },
            {
                "stage_name": "Stage 6",
                "commands": [
                    ("min_style", "cg"),
                    ("fix", "1 all box/relax iso 0.0 vmax 0.001"),
                    ("minimize", "1.0e-16 1.0e-16 5000 10000"),
                    ("write_data", "run.data"),
                ],
            },
        ]

    def test_append(self):
        lmp_input = LammpsInputFile().from_file(self.filename)
        lmp_input2 = LammpsInputFile().from_file(self.filename)
        lmp_input.append(lmp_input2)
        assert lmp_input.nstages == 2
        assert lmp_input.ncomments == 6
        assert lmp_input.stages_names == ["Stage 1", "Stage 2"]

        lmp_input = LammpsInputFile().from_file(self.filename, keep_stages=True)
        lmp_input2 = LammpsInputFile().from_file(self.filename, keep_stages=True)
        lmp_input.append(lmp_input2)
        assert lmp_input.nstages == 12
        assert lmp_input.ncomments == 2
        assert lmp_input.stages_names == [
            "Comment 1",
            "1) Initialization",
            "2) System definition",
            "3) Simulation settings",
            "Part A : energy minimization",
            "Stage 6",
            "Comment 2",
            "Stage 8 (previously 1) Initialization)",
            "Stage 9 (previously 2) System definition)",
            "Stage 10 (previously 3) Simulation settings)",
            "Stage 11 (previously Part A : energy minimization)",
            "Stage 12",
        ]

    def test_add_command(self):
        lmp_input = LammpsInputFile()
        lmp_input.stages.append({"stage_name": "Init", "commands": [("units", "metal")]})
        lmp_input._add_command(stage_name="Init", command="boundary p p p")
        assert lmp_input.stages == [
            {
                "stage_name": "Init",
                "commands": [("units", "metal"), ("boundary", "p p p")],
            }
        ]
        lmp_input._add_command(stage_name="Init", command="atom_style", args="full")
        assert lmp_input.stages == [
            {
                "stage_name": "Init",
                "commands": [
                    ("units", "metal"),
                    ("boundary", "p p p"),
                    ("atom_style", "full"),
                ],
            }
        ]

    def test_add_commands(self):
        lmp_input = LammpsInputFile()
        lmp_input.add_stage({"stage_name": "Init", "commands": [("units", "metal")]})
        lmp_input.add_commands(stage_name="Init", commands="boundary p p p")
        assert lmp_input.stages == [
            {
                "stage_name": "Init",
                "commands": [("units", "metal"), ("boundary", "p p p")],
            }
        ]
        lmp_input.add_commands(stage_name="Init", commands=["atom_style full", "dimension 3"])
        assert lmp_input.stages == [
            {
                "stage_name": "Init",
                "commands": [
                    ("units", "metal"),
                    ("boundary", "p p p"),
                    ("atom_style", "full"),
                    ("dimension", "3"),
                ],
            }
        ]
        lmp_input.add_commands(stage_name="Init", commands={"kspace_style": "ewald 1e-4"})
        assert lmp_input.stages == [
            {
                "stage_name": "Init",
                "commands": [
                    ("units", "metal"),
                    ("boundary", "p p p"),
                    ("atom_style", "full"),
                    ("dimension", "3"),
                    ("kspace_style", "ewald 1e-4"),
                ],
            }
        ]

    def test_add_comment(self):
        lmp_input = LammpsInputFile()
        lmp_input._add_comment(comment="First comment")
        assert lmp_input.stages == [{"stage_name": "Comment 1", "commands": [("#", "First comment")]}]
        lmp_input._add_comment(comment="Sub comment", stage_name="Comment 1")
        assert lmp_input.stages == [
            {
                "stage_name": "Comment 1",
                "commands": [("#", "First comment"), ("#", "Sub comment")],
            }
        ]
        lmp_input.stages.append({"stage_name": "Init", "commands": [("units", "metal")]})
        lmp_input._add_comment(comment="Inline comment", stage_name="Init")
        assert lmp_input.stages == [
            {
                "stage_name": "Comment 1",
                "commands": [("#", "First comment"), ("#", "Sub comment")],
            },
            {
                "stage_name": "Init",
                "commands": [("units", "metal"), ("#", "Inline comment")],
            },
        ]


class TestFunc(MatSciTest):
    @pytest.mark.filterwarnings("ignore:write_lammps_inputs is deprecated")
    def test_write_lammps_inputs(self):
        # script template
        with open(f"{TEST_DIR}/kappa.txt", encoding="utf-8") as file:
            kappa_template = file.read()
        kappa_settings = {"method": "heat"}
        write_lammps_inputs(output_dir="heat", script_template=kappa_template, settings=kappa_settings)
        with open(f"{self.tmp_path}/heat/in.lammps", encoding="utf-8") as file:
            kappa_script = file.read()
        fix_hot = re.search(r"fix\s+hot\s+all\s+([^\s]+)\s+", kappa_script)
        # placeholders supposed to be filled
        assert fix_hot.group(1) == "heat"
        fix_cold = re.search(r"fix\s+cold\s+all\s+([^\s]+)\s+", kappa_script)
        assert fix_cold.group(1) == "heat"
        lattice = re.search(r"lattice\s+fcc\s+(.*)\n", kappa_script)
        # parentheses not supposed to be filled
        assert lattice.group(1) == "${rho}"
        pair_style = re.search(r"pair_style\slj/cut\s+(.*)\n", kappa_script)
        assert pair_style.group(1) == "${rc}"

        with open(f"{TEST_DIR}/in.peptide", encoding="utf-8") as file:
            peptide_script = file.read()
        # copy data file
        src = f"{TEST_DIR}/data.quartz"
        write_lammps_inputs(output_dir="path", script_template=peptide_script, data=src)
        dst = f"{self.tmp_path}/path/data.peptide"
        assert filecmp.cmp(src, dst, shallow=False)
        # write data file from obj
        obj = LammpsData.from_file(src, atom_style="atomic")
        with pytest.warns(FutureWarning):
            write_lammps_inputs(output_dir="obj", script_template=peptide_script, data=obj)
        obj_read = LammpsData.from_file(f"{self.tmp_path}/obj/data.peptide", atom_style="atomic")
        pd.testing.assert_frame_equal(obj_read.masses, obj.masses)
        pd.testing.assert_frame_equal(obj_read.atoms, obj.atoms)


class TestLammpsTemplateGen(MatSciTest):
    def test_write_inputs(self):
        # simple script without data file
        lis = LammpsTemplateGen().get_input_set(
            script_template=f"{TEST_DIR}/kappa.txt",
            settings={"method": "heat"},
            data=None,
            data_filename="data.peptide",
        )
        assert len(lis) == 1
        lis.write_input(self.tmp_path / "heat")

        with open(self.tmp_path / "heat" / "in.lammps", encoding="utf-8") as file:
            kappa_script = file.read()
        fix_hot = re.search(r"fix\s+hot\s+all\s+([^\s]+)\s+", kappa_script)
        # placeholders supposed to be filled
        assert fix_hot.group(1) == "heat"
        fix_cold = re.search(r"fix\s+cold\s+all\s+([^\s]+)\s+", kappa_script)
        assert fix_cold.group(1) == "heat"
        lattice = re.search(r"lattice\s+fcc\s+(.*)\n", kappa_script)
        # parentheses not supposed to be filled
        assert lattice.group(1) == "${rho}"
        pair_style = re.search(r"pair_style\slj/cut\s+(.*)\n", kappa_script)
        assert pair_style.group(1) == "${rc}"

        # script with data file
        obj = LammpsData.from_file(f"{TEST_DIR}/data.quartz", atom_style="atomic")
        lis = LammpsTemplateGen().get_input_set(
            script_template=f"{TEST_DIR}/in.peptide",
            settings=None,
            data=obj,
            data_filename="data.peptide",
        )
        assert len(lis) == 2
        assert isinstance(lis["data.peptide"], LammpsData)
        lis.write_input(self.tmp_path / "obj")

        obj_read = LammpsData.from_file(str(self.tmp_path / "obj" / "data.peptide"), atom_style="atomic")
        pd.testing.assert_frame_equal(obj_read.masses, obj.masses)
        pd.testing.assert_frame_equal(obj_read.atoms, obj.atoms)



"""Support for Abinit input variables."""

from __future__ import annotations

import collections
import collections.abc
import string
from collections.abc import Sequence

import numpy as np

from pymatgen.core.units import bohr_to_angstrom, eV_to_Ha

_SPECIAL_DATASET_INDICES = (":", "+", "?")
_DATASET_INDICES = "".join([*string.digits, *_SPECIAL_DATASET_INDICES])
_UNITS = {
    "bohr": 1.0,
    "angstrom": 1.0 / bohr_to_angstrom,
    "hartree": 1.0,
    "Ha": 1.0,
    "eV": eV_to_Ha,
}


class InputVariable:
    """An Abinit input variable."""

    def __init__(self, name: str, value, units: str = "", valperline: int = 3) -> None:
        """Initialize an InputVariable.

        Args:
            name: Name of the variable.
            value: Value of the variable.
            units: String specifying one of the units supported by Abinit. Default: atomic units.
            valperline: Number of items printed per line.
        """
        self._name = name
        self.value = value
        self._units = units

        self.valperline = valperline  # Maximum number of values per line.
        if name == "bdgw":
            self.valperline = 2

        if isinstance(self.value, Sequence) and isinstance(self.value[-1], str) and self.value[-1] in _UNITS:
            self.value = list(self.value)
            self._units = self.value.pop(-1)

    def get_value(self):
        """Return the value."""
        if self.units:
            return [*self.value, self.units]
        return self.value

    @property
    def name(self):
        """Name of the variable."""
        return self._name

    @property
    def basename(self):
        """The name trimmed of any dataset index."""
        basename = self.name
        return basename.rstrip(_DATASET_INDICES)

    @property
    def dataset(self):
        """The dataset index in string form."""
        return self.name.split(self.basename)[-1]

    @property
    def units(self):
        """The units."""
        return self._units

    def __str__(self):
        """Declaration of the variable in the input file."""
        value = self.value
        if value is None or not str(value):
            return ""

        var = self.name
        line = " " + var

        # By default, do not impose a number of decimal points
        float_decimal = 0

        # For some inputs, enforce number of decimal points...
        if any(inp in var for inp in ("xred", "xcart", "rprim", "qpt", "kpt")):
            float_decimal = 16

        # ...but not for those
        if any(inp in var for inp in ("ngkpt", "kptrlatt", "ngqpt", "ng2qpt")):
            float_decimal = 0

        if isinstance(value, np.ndarray):
            value = list(value.flatten())

        # values in lists
        if isinstance(value, list | tuple):
            # Reshape a list of lists into a single list
            if all(isinstance(v, list | tuple) for v in value):
                line += self.format_list2d(value, float_decimal)

            else:
                line += self.format_list(value, float_decimal)

        # scalar values
        else:
            line += f" {value}"

        # Add units
        if self.units:
            line += f" {self.units}"

        return line

    @staticmethod
    def format_scalar(val, float_decimal=0):
        """Format a single numerical value into a string
        with the appropriate number of decimal.
        """
        str_val = str(val)
        if str_val.lstrip("-").lstrip("+").isdigit() and float_decimal == 0:
            return str_val

        try:
            fval = float(val)
        except Exception:
            return str_val

        if fval == 0 or (1e-3 < abs(fval) < 1e4):
            form = "f"
            add_len = 5
        else:
            form = "e"
            add_len = 8

        n_dec = max(len(str(fval - int(fval))) - 2, float_decimal)
        n_dec = min(n_dec, 10)

        str_val = f"{fval:>{n_dec + add_len}.{n_dec}{form}}"

        return str_val.replace("e", "d")

    @staticmethod
    def format_list2d(values, float_decimal=0):
        """Format a list of lists."""
        flattened_list = flatten(values)

        # Determine the representation
        if all(isinstance(v, int) for v in flattened_list):
            type_all = int
        else:
            try:
                for v in flattened_list:
                    float(v)
                type_all = float
            except Exception:
                type_all = str

        # Determine the format
        width = max(len(str(s)) for s in flattened_list)
        if type_all is int:
            fmt_spec = f">{width}d"
        elif type_all is str:
            fmt_spec = f">{width}"
        else:
            # Number of decimal
            max_dec = max(len(str(f - int(f))) - 2 for f in flattened_list)
            n_dec = min(max(max_dec, float_decimal), 10)

            if all(f == 0 or (abs(f) > 1e-3 and abs(f) < 1e4) for f in flattened_list):
                fmt_spec = f">{n_dec + 5}.{n_dec}f"
            else:
                fmt_spec = f">{n_dec + 8}.{n_dec}e"

        line = "\n"
        for lst in values:
            for val in lst:
                line += f" {val:{ {fmt_spec} }}"
            line += "\n"

        return line.rstrip("\n")

    def format_list(self, values, float_decimal=0):
        """Format a list of values into a string.
        The result might be spread among several lines.
        """
        line = ""

        # Format the line declaring the value
        for i, val in enumerate(values, start=1):
            line += f" {self.format_scalar(val, float_decimal)}"
            if self.valperline is not None and i % self.valperline == 0:
                line += "\n"

        # Add a carriage return in case of several lines
        if "\n" in line.rstrip("\n"):
            line = "\n" + line

        return line.rstrip("\n")


def flatten(iterable):
    """Make an iterable flat, i.e. a 1d iterable object."""
    iterator = iter(iterable)
    array, stack = collections.deque(), collections.deque()
    while True:
        try:
            value = next(iterator)
        except StopIteration:
            if not stack:
                return tuple(array)
            iterator = stack.pop()
        else:
            if not isinstance(value, str) and isinstance(value, collections.abc.Iterable):
                stack.append(iterator)
                iterator = iter(value)
            else:
                array.append(value)



"""Parity tests for pymatgen.core.constants vs scipy.constants.

pymatgen.core.constants embeds CODATA values as Python literals to avoid the
~50 ms import cost of scipy.constants. These tests assert byte-exact agreement
with the currently installed scipy. When scipy bumps to a new CODATA release,
these tests will fail and prompt an update of the embedded literals.
"""

from __future__ import annotations

import pytest
import scipy.constants as sc

from pymatgen.core import constants as pmg_const

# Top-level named constants we expose. Each is compared to the same-named
# attribute on scipy.constants.
NAMED_CONSTANTS = [
    "e",
    "N_A",
    "Avogadro",
    "Boltzmann",
    "k",
    "R",
    "hbar",
    "h",
    "c",
    "m_e",
    "epsilon_0",
    "mile",
    "calorie",
    "tera",
    "milli",
    "centi",
]


@pytest.mark.parametrize("name", NAMED_CONSTANTS)
def test_named_constant_matches_scipy(name: str) -> None:
    pmg_val = getattr(pmg_const, name)
    scipy_val = getattr(sc, name)
    assert pmg_val == scipy_val, f"{name}: pymatgen={pmg_val!r}  scipy={scipy_val!r}"


@pytest.mark.parametrize("key", list(pmg_const.physical_constants))
def test_physical_constant_matches_scipy(key: str) -> None:
    """Every key in pymatgen's physical_constants must match scipy exactly."""
    pmg_tuple = pmg_const.physical_constants[key]
    scipy_tuple = sc.physical_constants[key]
    assert pmg_tuple == scipy_tuple, f"{key}: pymatgen={pmg_tuple!r}  scipy={scipy_tuple!r}"


@pytest.mark.parametrize("key", list(pmg_const.physical_constants))
def test_value_helper_matches_scipy(key: str) -> None:
    """``pmg_const.value(key)`` must match ``scipy.constants.value(key)``."""
    assert pmg_const.value(key) == sc.value(key), f"value({key!r}) mismatch"


def test_no_scipy_constants_on_import_path() -> None:
    """Importing pymatgen.core.constants must not pull in scipy.constants.

    The whole point of this module is to avoid the scipy.constants subtree
    (which drags in numpy.f2py / fileinput / charset_normalizer). This test
    re-imports both in a fresh sys.modules-pruned subprocess to verify.
    """
    import subprocess
    import sys

    code = (
        "import sys; "
        "from pymatgen.core import constants; "
        "assert 'scipy.constants' not in sys.modules, "
        "    f'scipy.constants leaked: {sorted(m for m in sys.modules if m.startswith(\"scipy\"))!r}'; "
        "print('OK')"
    )
    result = subprocess.run([sys.executable, "-c", code], capture_output=True, text=True, check=False)
    assert result.returncode == 0, result.stderr
    assert "OK" in result.stdout



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
import pytest
from matplotlib.figure import Figure as MplFigure

from pymatgen.util.plotting import (
    add_fig_kwargs,
    format_formula,
    get_ax3d_fig,
    get_ax_fig,
    get_axarray_fig_plt,
    periodic_table_heatmap,
    pretty_plot,
    pretty_plot_two_axis,
    pretty_polyfit_plot,
    van_arkel_triangle,
)
from pymatgen.util.testing import MatSciTest

try:
    import pymatviz
    from plotly.graph_objects import Figure
except ImportError:
    pymatviz = None


class TestFunc(MatSciTest):
    def test_plot_periodic_heatmap(self):
        random_data = {"Te": 0.11083, "Au": 0.75756, "Th": 1.24758, "Ni": -2.0354}
        fig = periodic_table_heatmap(random_data)
        if pymatviz is not None:
            assert isinstance(fig, Figure)
        else:
            assert isinstance(fig, plt.Axes)

        # Test all keywords
        periodic_table_heatmap(
            random_data,
            cmap="plasma",
            max_row=10,
            cbar_label_size=18,
            cmap_range=[0, 1],
            cbar_label="Hello World",
            blank_color="white",
            value_format=".4f",
            edge_color="black",
            value_fontsize=12,
            symbol_fontsize=18,
            readable_fontcolor=True,
        )

    def test_van_arkel_triangle(self):
        random_list = [("Fe", "C"), ("Ni", "F")]
        ax = van_arkel_triangle(random_list)
        assert isinstance(ax, plt.Axes)
        assert ax.get_title() == ""
        assert ax.get_xlabel() == r"$\frac{\chi_{A}+\chi_{B}}{2}$"
        assert ax.get_ylabel() == r"$|\chi_{A}-\chi_{B}|$"
        ax = van_arkel_triangle(random_list, annotate=True)
        assert isinstance(ax, plt.Axes)


class TestPrettyPlot:
    def teardown_method(self):
        plt.close("all")

    def test_default_returns_axes(self):
        ax = pretty_plot()
        assert isinstance(ax, plt.Axes)

    def test_height_defaults_to_golden_ratio(self):
        ax = pretty_plot(width=10)
        fig = ax.figure
        w, h = fig.get_size_inches()
        assert w == 10
        # int(width * golden_ratio) where golden = (sqrt(5)-1)/2 ≈ 0.618
        assert h == int(10 * ((5**0.5 - 1) / 2))

    def test_with_existing_axes_resizes_figure(self):
        fig, ax = plt.subplots()
        out = pretty_plot(width=6, height=4, ax=ax)
        assert out is ax
        assert tuple(fig.get_size_inches()) == (6, 4)


class TestPrettyPlotTwoAxis:
    def teardown_method(self):
        plt.close("all")

    def test_sequence_inputs(self):
        x = [0, 1, 2, 3]
        y1 = [1, 2, 3, 4]
        y2 = [4, 3, 2, 1]
        ax1 = pretty_plot_two_axis(x, y1, y2, xlabel="x", y1label="y1", y2label="y2", dpi=None)
        assert isinstance(ax1, plt.Axes)
        assert ax1.get_xlabel() == "x"
        assert ax1.get_ylabel() == "y1"

    def test_dict_inputs(self):
        x = [0, 1, 2]
        y1 = {"a": [1, 2, 3], "b": [3, 2, 1]}
        y2 = {"c": [0, 1, 0]}
        ax1 = pretty_plot_two_axis(x, y1, y2, dpi=None)
        assert isinstance(ax1, plt.Axes)


class TestPrettyPolyfitPlot:
    def teardown_method(self):
        plt.close("all")

    def test_linear_fit(self):
        x = np.array([1.0, 2.0, 3.0, 4.0])
        y = 2 * x + 1
        ax = pretty_polyfit_plot(x, y, deg=1, xlabel="x", ylabel="y")
        assert isinstance(ax, plt.Axes)
        assert ax.get_xlabel() == "x"
        assert ax.get_ylabel() == "y"

    def test_quadratic_fit(self):
        x = np.linspace(-2, 2, 20)
        y = x**2
        ax = pretty_polyfit_plot(x, y, deg=2)
        assert isinstance(ax, plt.Axes)


class TestFormatFormula:
    @pytest.mark.parametrize(
        ("formula", "expected"),
        [
            ("Li2O", "$Li_{2}O$"),
            ("Fe", "$Fe$"),
            ("LiFePO4", "$LiFePO_{4}$"),
            ("H2O", "$H_{2}O$"),
            ("Li3V2(PO4)3", "$Li_{3}V_{2}(PO_{4})_{3}$"),
        ],
    )
    def test_format(self, formula, expected):
        assert format_formula(formula) == expected


class TestAxFigHelpers:
    def teardown_method(self):
        plt.close("all")

    def test_get_ax_fig_creates_when_none(self):
        ax, fig = get_ax_fig()
        assert isinstance(ax, plt.Axes)
        assert isinstance(fig, MplFigure)

    def test_get_ax_fig_passes_through(self):
        existing_ax = plt.figure().gca()
        ax, _fig = get_ax_fig(ax=existing_ax)
        assert ax is existing_ax

    def test_get_ax3d_fig_creates_when_none(self):
        ax, fig = get_ax3d_fig()
        assert isinstance(fig, MplFigure)
        assert ax.name == "3d"

    def test_get_ax3d_fig_passes_through(self):
        existing_ax = plt.figure().add_subplot(projection="3d")
        ax, _fig = get_ax3d_fig(ax=existing_ax)
        assert ax is existing_ax

    def test_get_axarray_fig_plt_creates(self):
        ax_array, fig, plt_mod = get_axarray_fig_plt(None, nrows=2, ncols=2)
        assert ax_array.shape == (2, 2)
        assert isinstance(fig, MplFigure)
        assert plt_mod is plt

    def test_get_axarray_fig_plt_squeezes_single(self):
        ax_array, _fig, _ = get_axarray_fig_plt(None, nrows=1, ncols=1)
        assert isinstance(ax_array, plt.Axes)

    def test_get_axarray_fig_plt_passthrough(self):
        _fig, axes = plt.subplots(1, 2)
        ax_array, _fig_out, _ = get_axarray_fig_plt(axes, nrows=1, ncols=2)
        assert ax_array.shape == (2,)


class TestAddFigKwargs:
    def teardown_method(self):
        plt.close("all")

    def _make_fn(self):
        @add_fig_kwargs
        def make_fig():
            fig, ax = plt.subplots()
            ax.plot([0, 1], [0, 1])
            return fig

        return make_fig

    def test_returns_figure_with_title(self):
        fig = self._make_fn()(title="hello", show=False)
        assert isinstance(fig, MplFigure)
        assert fig._suptitle.get_text() == "hello"

    def test_size_kwargs(self):
        fig = self._make_fn()(size_kwargs={"w": 6, "h": 4}, show=False)
        assert tuple(fig.get_size_inches()) == (6, 4)

    def test_savefig(self, tmp_path):
        out = tmp_path / "out.png"
        self._make_fn()(savefig=str(out), show=False)
        assert out.exists()

    def test_ax_grid_and_tight_layout(self):
        fig = self._make_fn()(ax_grid=True, tight_layout=True, show=False)
        assert all(any(line.get_visible() for line in ax.get_xgridlines()) for ax in fig.axes)

    def test_ax_annotate(self):
        @add_fig_kwargs
        def make_two_panel():
            fig, _ = plt.subplots(1, 2)
            return fig

        fig = make_two_panel(ax_annotate=True, show=False)
        # Each axis should have one annotation: "(a)", "(b)"
        for ax, expected in zip(fig.axes, ["(a)", "(b)"], strict=True):
            assert any(child.get_text() == expected for child in ax.get_children() if hasattr(child, "get_text"))

    def test_fig_close_returns_fig(self):
        fig = self._make_fn()(show=False, fig_close=True)
        assert isinstance(fig, MplFigure)

    def test_passthrough_when_func_returns_none(self):
        @add_fig_kwargs
        def make_nothing():
            return None

        assert make_nothing(title="x", show=False) is None



"""This module implements an interface to the Henkelmann et al.'s excellent
Fortran code for calculating a Bader charge analysis.

This module depends on a compiled bader executable available in the path.
Please download the library at https://theory.cm.utexas.edu/henkelman/code/bader/
and follow the instructions to compile the executable.

If you use this module, please cite:

G. Henkelman, A. Arnaldsson, and H. Jonsson, "A fast and robust algorithm for
Bader decomposition of charge density", Comput. Mater. Sci. 36, 254-360 (2006).
"""

from __future__ import annotations

import contextlib
import shutil
import subprocess
import warnings
from glob import glob
from pathlib import Path
from shutil import which
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING

import numpy as np
from monty.shutil import decompress_file
from monty.tempfile import ScratchDir

from pymatgen.core import Structure
from pymatgen.io.common import VolumetricData
from pymatgen.io.vasp.inputs import Potcar
from pymatgen.io.vasp.outputs import Chgcar

if TYPE_CHECKING:
    from typing import Any, Self

    from pymatgen.core import IStructure

__author__ = "shyuepingong"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyuep@gmail.com"
__status__ = "Beta"
__date__ = "4/5/13"


class BaderAnalysis:
    """Perform Bader charge analysis for Cube files or VASP outputs.

    Attributes:
        data (list[dict]): Atomic data parsed from bader analysis.
            Each dictionary in the list has the keys:
            "atomic_vol", "min_dist", "charge", "x", "y", "z".
        vacuum_volume (float): Vacuum volume of the Bader analysis.
        vacuum_charge (float): Vacuum charge of the Bader analysis.
        nelectrons (int): Number of electrons of the Bader analysis.
        chgcar (Chgcar): Chgcar object associated with input CHGCAR file.
    """

    def __init__(
        self,
        chgcar_filename: str = "",
        potcar_filename: str = "",
        chgref_filename: str = "",
        cube_filename: str = "",
        bader_path: str | None = None,
        parse_atomic_densities: bool | None = None,
    ) -> None:
        """Initialize the Bader caller.

        Args:
            chgcar_filename (str): The filename of the CHGCAR.
            potcar_filename (str): The filename of the POTCAR.
            chgref_filename (str): The filename of the
                reference charge density.
            cube_filename (str, optional): The filename of the cube file.
            bader_path (str, optional): The path to the bader executable.
            parse_atomic_densities (bool | None, optional): Deprecated and has no effect.
                It will be removed on 2026-12-31.
        """
        if parse_atomic_densities is not None:
            warnings.warn(
                "parse_atomic_densities is deprecated and will be removed on 2026-12-31; it no longer has any effect.",
                DeprecationWarning,
                stacklevel=2,
            )

        def temp_decompress(file: str | Path, target_dir: str = ".") -> str:
            """Utility function to copy a compressed file to a target directory (ScratchDir)
            and decompress it, to avoid modifying files in place.

            Args:
                file (str | Path): The path to the compressed file to be decompressed.
                target_dir (str, optional): The target directory where the decompressed file will be stored.
                    Defaults to "." (current directory).

            Returns:
                str: The path to the decompressed file if successful.
            """
            file = Path(file)

            if file.suffix.lower() in {".bz2", ".gz", ".z"}:
                shutil.copy(file, f"{target_dir}/{file.name}")
                out_file = decompress_file(f"{target_dir}/{file.name}")

                if file:  # to avoid mypy error
                    return str(out_file)

                raise FileNotFoundError(f"File {out_file} not found.")

            return str(file)

        # Get and check Bader executable path
        if bader_path is None:
            bader_path = which("bader") or which("bader.exe")

        if bader_path is None:
            raise RuntimeError(
                "Requires bader or bader.exe to be in the PATH or the absolute path to "
                f"the binary to be specified via {bader_path=}.\n"
                "Download from https://theory.cm.utexas.edu/henkelman/code/bader."
            )

        # Check input cube/CHGCAR files
        if not (cube_filename or chgcar_filename):
            raise ValueError("You must provide either a Cube file or a CHGCAR.")
        if cube_filename and chgcar_filename:
            raise ValueError("Cannot parse cube and CHGCAR at the same time.")

        with ScratchDir("."):
            if chgcar_filename:
                filepath = chgcar_fpath = temp_decompress(chgcar_filename)

                self.chgcar = Chgcar.from_file(chgcar_fpath)
                self.structure = self.chgcar.structure

                self.potcar = Potcar.from_file(potcar_filename) if potcar_filename else None
                self.natoms = self.chgcar.poscar.natoms

                # List of nelects for each atom from potcar
                potcar_indices = []
                for idx, val in enumerate(self.natoms):
                    potcar_indices += [idx] * val
                self.nelects = (
                    [self.potcar[potcar_indices[i]].nelectrons for i in range(len(self.structure))]
                    if self.potcar
                    else []
                )

            # Parse from Cube file
            else:
                filepath = cube_fpath = temp_decompress(cube_filename)

                self.cube = VolumetricData.from_cube(cube_fpath)
                self.structure = self.cube.structure
                self.nelects = []

            # Prepare CHGCAR reference file
            chgref_fpath = temp_decompress(chgref_filename)
            self.reference_used = bool(chgref_filename)

            bader_args = [bader_path, filepath]

            if self.reference_used:
                bader_args += ["-ref", chgref_fpath]

            with subprocess.Popen(
                bader_args,
                stdout=subprocess.PIPE,
                stdin=subprocess.PIPE,
                close_fds=True,
            ) as proc:
                stdout, stderr = proc.communicate()
                if proc.returncode != 0:
                    raise RuntimeError(
                        f"{bader_path} exit code: {proc.returncode}, error: {stderr!s}."
                        f"\nstdout: {stdout!s}. Please check your bader installation."
                    )

            # Determine Bader version
            try:
                self.version = float(stdout.split()[5])
            except ValueError:
                self.version = -1  # Unknown

            if self.version < 1.0:
                warnings.warn(
                    "Your installed version of Bader is outdated, calculation of vacuum charge may be incorrect.",
                    stacklevel=2,
                )

            # Parse ACF.dat file
            self.data = self._parse_acf()

    def _parse_acf(self) -> list[dict]:
        """Parse Bader output file ACF.dat."""
        with open("ACF.dat", encoding="us-ascii") as file:
            lines = file.readlines()

        # Skip header lines
        headers = ("x", "y", "z", "charge", "min_dist", "atomic_vol")
        lines.pop(0)
        lines.pop(0)

        data = []
        while True:
            line = lines.pop(0).strip()
            if line.startswith("-"):
                break
            vals = map(float, line.split()[1:])
            data.append(dict(zip(headers, vals, strict=True)))

        for line in lines:
            tokens = line.strip().split(":")
            if tokens[0] == "VACUUM CHARGE":
                self.vacuum_charge = float(tokens[1])
            elif tokens[0] == "VACUUM VOLUME":
                self.vacuum_volume = float(tokens[1])
            elif tokens[0] == "NUMBER OF ELECTRONS":
                self.nelectrons = float(tokens[1])

        return data

    def get_charge(self, atom_index: int) -> float:
        """Convenience method to get the charge on a particular atom. This is the "raw"
        charge generated by the Bader program, not a partial atomic charge. If the cube file
        is a spin-density file, then this will return the spin density per atom with
        positive being spin up and negative being spin down.

        Args:
            atom_index (int): Index of atom.

        Returns:
            float: Charge associated with atom from the Bader analysis.
        """
        return self.data[atom_index]["charge"]

    def get_charge_transfer(self, atom_index: int, nelect: int | None = None) -> float:
        """Get the charge transferred for a particular atom. A positive value means
        that the site has gained electron density (i.e. exhibits anionic character)
        whereas a negative value means the site has lost electron density (i.e. exhibits
        cationic character). If the arg nelect is not supplied, then POTCAR must be
        supplied to determine nelect.

        Args:
            atom_index (int): Index of atom.
            nelect (int): number of electrons associated with an isolated atom at this index.
                For most DFT codes this corresponds to the number of valence electrons
                associated with the pseudopotential (e.g. ZVAL for VASP).

        Returns:
            Charge transfer associated with atom from the Bader analysis.
            Given by bader charge on atom - nelect for associated atom.
        """
        if not self.nelects and nelect is None:
            raise ValueError("No NELECT info! Need POTCAR for VASP or nelect argument for cube file")
        return self.data[atom_index]["charge"] - (nelect if nelect is not None else self.nelects[atom_index])

    def get_partial_charge(self, atom_index: int, nelect: int | None = None) -> float:
        """Convenience method to get the partial charge on a particular atom. This is
        simply the negative value of the charge transferred. A positive value indicates
        that the atom has cationic character, whereas a negative value indicates the
        site has anionic character.

        Args:
            atom_index (int): Index of atom.
            nelect (int): number of electrons associated with an isolated atom at this index.
                For most DFT codes this corresponds to the number of valence electrons
                associated with the pseudopotential (e.g. ZVAL for VASP).

        Returns:
            Charge associated with atom from the Bader analysis.
        """
        return -self.get_charge_transfer(atom_index, nelect)

    def get_charge_decorated_structure(self) -> Structure | IStructure:
        """Get a charge decorated structure.

        Note, this assumes that the Bader analysis was correctly performed on a file
        with electron densities
        """
        charges = [-self.get_charge(i) for i in range(len(self.structure))]
        struct = self.structure.copy()
        struct.add_site_property("charge", charges)
        return struct

    def get_oxidation_state_decorated_structure(self, nelects: list[int] | None = None) -> Structure:
        """Get an oxidation state decorated structure based on bader analysis results.
        Each site is assigned a charge based on the computed partial atomic charge from bader.

        Note, this assumes that the Bader analysis was correctly performed on a file
        with electron densities.

        Args:
            nelects (list[int]): number of electrons associated with an isolated atom at this index.

        Returns:
            Structure: with bader-analysis-based oxidation states.
        """
        struct = Structure.from_sites(self.structure)
        charges = [self.get_partial_charge(idx, None if not nelects else nelects[idx]) for idx in range(len(struct))]
        struct.add_oxidation_state_by_site(charges)
        return struct

    def get_decorated_structure(self, property_name: str, average: bool = False) -> Structure:
        """Get a property-decorated structure from the Bader analysis.

        This is distinct from getting charge decorated structure, which assumes
        the "standard" Bader analysis of electron densities followed by converting
        electron count to charge. The expected way to use this is to call Bader on
        a non-charge density file such as a spin density file, electrostatic potential
        file, etc., while using the charge density file as the reference (chgref_filename)
        so that the partitioning is determined via the charge, but averaging or integrating
        is done for another property.

        User warning: Bader analysis cannot automatically determine what property is
        inside of the file. So if you want to use this for a non-conventional property
        like spin, you must ensure that you have the file is for the appropriate
        property and you have an appropriate reference file.

        Args:
            property_name (str): name of the property to assign to the structure, note that
                if name is "spin" this is handled as a special case, and the appropriate
                spin properties are set on the species in the structure
            average (bool): whether or not to return the average of this property, rather
                than the total, by dividing by the atomic volume.

        Returns:
            structure with site properties assigned via Bader Analysis
        """
        vals = np.array([self.get_charge(i) for i in range(len(self.structure))])
        struct = Structure.from_sites(self.structure)
        if average:
            vals = np.divide(vals, [d["atomic_vol"] for d in self.data])
        struct.add_site_property(property_name, vals)
        if property_name == "spin":
            struct.add_spin_by_site(list(vals))
        return struct

    @property
    def summary(self) -> dict[str, Any]:
        """Dict summary of key analysis, e.g. atomic volume, charge, etc."""
        summary = {
            "min_dist": [d["min_dist"] for d in self.data],
            "charge": [d["charge"] for d in self.data],
            "atomic_volume": [d["atomic_vol"] for d in self.data],
            "vacuum_charge": self.vacuum_charge,
            "vacuum_volume": self.vacuum_volume,
            "reference_used": self.reference_used,
            "bader_version": self.version,
        }

        if self.potcar:
            charge_transfer = [self.get_charge_transfer(i) for i in range(len(self.data))]
            summary["charge_transfer"] = charge_transfer

        return summary

    @classmethod
    def from_path(cls, path: str, suffix: str = "") -> Self:
        """Convenient constructor that takes in the path name of VASP run
        to perform Bader analysis.

        Args:
            path (str): Name of directory where VASP output files are stored.
            suffix (str): specific suffix to look for (e.g. '.relax1' for 'CHGCAR.relax1.gz').

        Returns:
            BaderAnalysis
        """

        def _get_filepath(filename):
            name_pattern = f"{filename}{suffix}*" if filename != "POTCAR" else f"{filename}*"
            paths = glob(f"{path}/{name_pattern}")
            filepath = ""
            if len(paths) >= 1:
                # using reverse=True because, if multiple files are present, they likely
                # have suffixes 'static', 'relax', 'relax2', etc. and this would give
                # 'static' over 'relax2' over 'relax' however, better to use 'suffix'
                # kwarg to avoid this!
                paths.sort(reverse=True)
                if len(paths) > 1:
                    warnings.warn(f"Multiple files detected, using {paths[0]}", stacklevel=2)
                filepath = paths[0]
            else:
                msg = f"Could not find {filename!r}"
                if filename in ("AECCAR0", "AECCAR2"):
                    msg += ", interpret Bader results with severe caution."
                elif filename == "POTCAR":
                    msg += ", cannot calculate charge transfer."
                warnings.warn(msg, stacklevel=2)
            return filepath

        chgcar_filename = _get_filepath("CHGCAR")
        if chgcar_filename is None:
            raise FileNotFoundError("Could not find CHGCAR!")
        potcar_filename = _get_filepath("POTCAR")
        aeccar0 = _get_filepath("AECCAR0")
        aeccar2 = _get_filepath("AECCAR2")
        if aeccar0 and aeccar2:
            # `chgsum.pl AECCAR0 AECCAR2` equivalent to obtain chgref_file
            chgref = Chgcar.from_file(aeccar0) + Chgcar.from_file(aeccar2)
            chgref_filename = f"{path}/CHGCAR_ref"
            chgref.write_file(chgref_filename)
        else:
            chgref_filename = ""
        return cls(
            chgcar_filename=chgcar_filename,
            potcar_filename=potcar_filename,
            chgref_filename=chgref_filename,
        )


def bader_analysis_from_path(path: str, suffix: str = "") -> dict[str, Any]:
    """Convenience method to run Bader analysis on a folder containing
    typical VASP output files.

    This method will:

    1. Look for files CHGCAR, AECCAR0, AECCAR2, POTCAR or their gzipped
    counterparts.
    2. If AECCAR* files are present, constructs a temporary reference
    file as AECCAR0 + AECCAR2
    3. Runs Bader analysis twice: once for charge, and a second time
    for the charge difference (magnetization density).

    Args:
        path: path to folder to search in
        suffix: specific suffix to look for (e.g. '.relax1' for 'CHGCAR.relax1.gz'

    Returns:
        summary dict
    """

    def _get_filepath(filename: str, msg: str = "") -> str | None:
        paths = glob(glob_pattern := f"{path}/{filename}{suffix}*")
        if len(paths) == 0:
            warnings.warn(msg or f"no matches for {glob_pattern=}", stacklevel=2)
            return None
        if len(paths) > 1:
            # using reverse=True because, if multiple files are present,
            # they likely have suffixes 'static', 'relax', 'relax2', etc.
            # and this would give 'static' over 'relax2' over 'relax'
            # however, better to use 'suffix' kwarg to avoid this!
            paths.sort(reverse=True)
            warnings.warn(f"Multiple files detected, using {Path(path).name}", stacklevel=2)
        return paths[0]

    chgcar_path = _get_filepath("CHGCAR", "Could not find CHGCAR!")
    if chgcar_path is None:
        raise FileNotFoundError("Could not find CHGCAR!")
    chgcar = Chgcar.from_file(chgcar_path)

    aeccar0_path = _get_filepath("AECCAR0")
    if not aeccar0_path:
        warnings.warn("Could not find AECCAR0, interpret Bader results with severe caution!", stacklevel=2)
    aeccar0 = Chgcar.from_file(aeccar0_path) if aeccar0_path else None

    aeccar2_path = _get_filepath("AECCAR2")
    if not aeccar2_path:
        warnings.warn("Could not find AECCAR2, interpret Bader results with severe caution!", stacklevel=2)
    aeccar2 = Chgcar.from_file(aeccar2_path) if aeccar2_path else None

    potcar_path = _get_filepath("POTCAR")
    if not potcar_path:
        warnings.warn("Could not find POTCAR, cannot calculate charge transfer.", stacklevel=2)
    potcar = Potcar.from_file(potcar_path) if potcar_path else None

    return bader_analysis_from_objects(chgcar, potcar, aeccar0, aeccar2)


def bader_analysis_from_objects(
    chgcar: Chgcar,
    potcar: Potcar | None = None,
    aeccar0: Chgcar | None = None,
    aeccar2: Chgcar | None = None,
) -> dict[str, Any]:
    """Convenience method to run Bader analysis from a set
    of pymatgen Chgcar and Potcar objects.

    This method will:

    1. If aeccar objects are present, constructs a temporary reference
    file as AECCAR0 + AECCAR2
    2. Runs Bader analysis twice: once for charge, and a second time
    for the charge difference (magnetization density).

    Args:
        chgcar: Chgcar object
        potcar: (optional) Potcar object
        aeccar0: (optional) Chgcar object from aeccar0 file
        aeccar2: (optional) Chgcar object from aeccar2 file

    Returns:
        summary dict
    """
    with TemporaryDirectory() as tmp_dir, contextlib.chdir(tmp_dir):
        if aeccar0 and aeccar2:
            # construct reference file
            chgref = aeccar0.linear_add(aeccar2)
            chgref_path = f"{tmp_dir}/CHGCAR_ref"
            chgref.write_file(chgref_path)
        else:
            chgref_path = ""

        chgcar.write_file("CHGCAR")
        chgcar_path = f"{tmp_dir}/CHGCAR"

        if potcar:
            potcar.write_file("POTCAR")
            potcar_path = f"{tmp_dir}/POTCAR"
        else:
            potcar_path = ""

        ba = BaderAnalysis(
            chgcar_filename=chgcar_path,
            potcar_filename=potcar_path,
            chgref_filename=chgref_path,
        )

        summary = {
            "min_dist": [dct["min_dist"] for dct in ba.data],
            "charge": [dct["charge"] for dct in ba.data],
            "atomic_volume": [dct["atomic_vol"] for dct in ba.data],
            "vacuum_charge": ba.vacuum_charge,
            "vacuum_volume": ba.vacuum_volume,
            "reference_used": bool(chgref_path),
            "bader_version": ba.version,
        }

        if potcar:
            charge_transfer = [ba.get_charge_transfer(idx) for idx in range(len(ba.data))]
            summary["charge_transfer"] = charge_transfer

        if chgcar.is_spin_polarized:
            # write a CHGCAR containing magnetization density only
            chgcar.data["total"] = chgcar.data["diff"]
            chgcar.is_spin_polarized = False
            chgcar.write_file("CHGCAR_mag")

            chgcar_mag_path = f"{tmp_dir}/CHGCAR_mag"
            ba = BaderAnalysis(
                chgcar_filename=chgcar_mag_path,
                potcar_filename=potcar_path,
                chgref_filename=chgref_path,
            )
            summary["magmom"] = [dct["charge"] for dct in ba.data]

    return summary
